import functools
import hashlib
import http.server
import multiprocessing
import os
import random
import shutil
import socket
import subprocess
import sys
from contextlib import closing
from pathlib import Path

import psutil
import pytest

import salt.utils.files
from tests.support.runtests import RUNTIME_VARS

try:
    import gnupg as gnupglib

    HAS_GNUPG = True
except ImportError:
    HAS_GNUPG = False


class TestRequestHandler(http.server.SimpleHTTPRequestHandler):
    """
    Modified request handler class
    """

    def __init__(self, *args, directory=None, **kwargs):
        if directory is None:
            directory = os.getcwd()
        self.directory = directory
        if sys.version_info.minor < 7:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(*args, directory=directory, **kwargs)

    def do_GET(self):
        """
        GET request handling
        """
        none_match = self.headers.get("If-None-Match")
        status_code = 200
        try:
            # Retrieve the local file from the web root to serve to clients
            with salt.utils.files.fopen(
                os.path.join(self.directory, self.path[1:]), "rb"
            ) as reqfp:
                return_data = reqfp.read()
                # We're using this checksum as the etag to show file changes
                checksum = hashlib.sha256(return_data).hexdigest()
                if none_match == checksum:
                    # Status code 304 Not Modified is returned if the file is unchanged
                    status_code = 304
        except:  # pylint: disable=bare-except
            # Something went wrong. We didn't find the requested file
            status_code = 404
            return_data = None
            checksum = None

        self.send_response(status_code)

        # Return the Etag header if we have the checksum
        if checksum:
            # IMPORTANT: This introduces randomness into the tests. The Etag header key
            # will be converted to lowercase in the code... but if someone breaks that,
            # it'll rear it's head here as random failures that are hard to reproduce.
            # Any alternatives seem overly complex. So... don't break the case insensitivity
            # in the code.
            possible_etags = ["Etag", "ETag", "etag", "ETAG"]
            self.send_header(random.choice(possible_etags), checksum)
            self.end_headers()

        # Return file content
        if return_data:
            self.wfile.write(return_data)


def serve(port=8000, directory=None):
    """
    Function to serve a directory via http.server
    """
    handler = functools.partial(TestRequestHandler, directory=directory)
    s = http.server.HTTPServer(("127.0.0.1", port), handler)
    s.serve_forever()


@pytest.fixture(scope="module")
def free_port():
    """
    Utility fixture to grab a free port for the web server
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


@pytest.fixture(autouse=True, scope="module")
def server(free_port, web_root):
    """
    Web server fixture
    """
    p = multiprocessing.Process(target=serve, args=(free_port, web_root))
    p.start()
    yield
    p.terminate()
    p.join()


@pytest.fixture(scope="module")
def web_root(tmp_path_factory):
    """
    Temporary directory fixture for the web server root
    """
    _web_root = tmp_path_factory.mktemp("web_root")
    try:
        yield str(_web_root)
    finally:
        shutil.rmtree(str(_web_root), ignore_errors=True)


@pytest.mark.slow_test
def test_archive_extracted_web_source_etag_operation(
    modules, states, free_port, web_root, minion_opts
):
    """
    This functional test checks the operation of the use_etag parameter to the
    archive.extracted state. There are four (4) invocations of archive.extracted
    with a web source, but only three (3) will trigger a call to the web server
    as shown below and in comments within.

        127.0.0.1 - - [08/Mar/2022 13:07:10] "GET /foo.tar.gz HTTP/1.1" 200 -
        127.0.0.1 - - [08/Mar/2022 13:07:10] "GET /foo.tar.gz HTTP/1.1" 304 -
        127.0.0.1 - - [08/Mar/2022 13:07:10] "GET /foo.tar.gz HTTP/1.1" 200 -

    Checks are documented in the comments.
    """
    # Create file in the web root directory to serve
    states.file.managed(
        name=os.path.join(web_root, "foo", "bar.txt"),
        contents="this is my file",
        makedirs=True,
    )
    modules.archive.tar(
        options="czf",
        tarfile=os.path.join(web_root, "foo.tar.gz"),
        sources=[os.path.join(web_root, "foo")],
        cwd=web_root,
    )

    # File should not be cached yet
    cached_file = os.path.join(
        minion_opts["cachedir"],
        "extrn_files",
        "base",
        f"localhost{free_port}",
        "foo.tar.gz",
    )
    cached_etag = cached_file + ".etag"
    assert not os.path.exists(cached_file)
    assert not os.path.exists(cached_etag)

    # Pull the file from the web server
    #     Web server returns 200 status code with content:
    #     127.0.0.1 - - [08/Mar/2022 13:07:10] "GET /foo.tar.gz HTTP/1.1" 200 -
    states.archive.extracted(
        name=web_root,
        source=f"http://localhost:{free_port}/foo.tar.gz",
        archive_format="tar",
        options="z",
        use_etag=True,
    )

    # Now the file is cached
    assert os.path.exists(cached_file)
    assert os.path.exists(cached_etag)

    # Store the original modified time of the cached file
    cached_file_mtime = os.path.getmtime(cached_file)

    # Pull the file again. Etag hasn't changed. No download occurs.
    #     Web server returns 304 status code and no content:
    #     127.0.0.1 - - [08/Mar/2022 13:07:10] "GET /foo.tar.gz HTTP/1.1" 304 -
    states.archive.extracted(
        name=web_root,
        source=f"http://localhost:{free_port}/foo.tar.gz",
        archive_format="tar",
        options="z",
        use_etag=True,
    )

    # Check that the modified time of the cached file hasn't changed
    assert cached_file_mtime == os.path.getmtime(cached_file)

    # Change file in the web root directory
    states.file.managed(
        name=os.path.join(web_root, "foo", "bar.txt"),
        contents="this is my changed file",
    )
    modules.archive.tar(
        options="czf",
        tarfile=os.path.join(web_root, "foo.tar.gz"),
        sources=[os.path.join(web_root, "foo")],
        cwd=web_root,
    )

    # Don't use Etag. Cached file is there, Salt won't try to download.
    #     No call to the web server will be made.
    states.archive.extracted(
        name=web_root,
        source=f"http://localhost:{free_port}/foo.tar.gz",
        archive_format="tar",
        options="z",
        use_etag=False,
    )

    # Check that the modified time of the cached file hasn't changed
    assert cached_file_mtime == os.path.getmtime(cached_file)

    # Now use Etag again. Cached file changes
    #     Web server returns 200 status code with content
    #     127.0.0.1 - - [08/Mar/2022 13:07:10] "GET /foo.tar.gz HTTP/1.1" 200 -
    states.archive.extracted(
        name=web_root,
        source=f"http://localhost:{free_port}/foo.tar.gz",
        archive_format="tar",
        options="z",
        use_etag=True,
    )

    # The modified time of the cached file now changes
    assert cached_file_mtime != os.path.getmtime(cached_file)


@pytest.fixture
def gpghome(tmp_path):
    root = tmp_path / "gpghome"
    root.mkdir(mode=0o0700)
    try:
        yield root
    finally:
        # Make sure we don't leave any gpg-agents running behind
        gpg_connect_agent = shutil.which("gpg-connect-agent")
        if gpg_connect_agent:
            gnupghome = root / ".gnupg"
            if not gnupghome.is_dir():
                gnupghome = root
            try:
                subprocess.run(
                    [gpg_connect_agent, "killagent", "/bye"],
                    env={"GNUPGHOME": str(gnupghome)},
                    shell=False,
                    check=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
            except subprocess.CalledProcessError:
                # This is likely CentOS 7 or Amazon Linux 2
                pass

        # If the above errored or was not enough, as a last resort, let's check
        # the running processes.
        for proc in psutil.process_iter():
            try:
                if "gpg-agent" in proc.name():
                    for arg in proc.cmdline():
                        if str(root) in arg:
                            proc.terminate()
            except Exception:  # pylint: disable=broad-except
                pass


@pytest.fixture
def gnupg(gpghome):
    return gnupglib.GPG(gnupghome=str(gpghome))


@pytest.fixture
def a_pubkey():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY9pxawEEAPpBbXxRYFUm6np5h746Nch7+OrbLdtBxP8x7VDOockr/x7drssb
llVFuK4HmiJg+Nkyakn3XmVYBHY2yBIkN/MP+R1zRxiFmniKOTD15UuHSQaWZTqh
qac6XrLZ20BiWl1fKweCz1wGUcMZaOBs0WVB0sIupqfS90Ub93VC/+oxABEBAAG0
JlNhbHRTdGFjayBBIFRlc3QgPGF0ZXN0QHNhbHRzdGFjay5jb20+iNEEEwEIADsW
IQT4zMjLXh2IaNqlhZqx+apXxJfnHAUCY9pxawIbAwULCQgHAgIiAgYVCgkICwIE
FgIDAQIeBwIXgAAKCRCx+apXxJfnHFDhA/47t5yYdCcjxXu/1Kn9sQwI+aq/S3x9
/ZKE+RodlryqA43BUT7N6JLQ5zJO6p+kRhMwCcVfBeDNJANqVi63HEDp8q3633BF
q1Cbi3BG0ugBdCADIETYBwl/ytMSgYwRO8b4TkYCyhWuWAgliVF3ceX0AVsng8pF
o6Vh4A3SqosQgA==
=eHpb
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def a_fp():
    return "F8CCC8CB5E1D8868DAA5859AB1F9AA57C497E71C"


@pytest.fixture
def b_pubkey():
    return """\
-----BEGIN PGP PUBLIC KEY BLOCK-----

mI0EY9pxigEEANbHCh566IEbp9Ez1WE3oEi+XXyf7H3GDgrVc8v9COMexpAFkJa1
gG+yCm4bOZ5vHAXbP2rGvlOEcao3y3evj2TWahg0+05CDugRjL0pO4JcMUBV1mBZ
ynUGoQ5T+WtKilJ5k/JrSRpJW3y//46q0g5c470qVNn9ZX0YZW/b7DFXABEBAAG0
JlNhbHRTdGFjayBCIFRlc3QgPGJ0ZXN0QHNhbHRzdGFjay5jb20+iNEEEwEIADsW
IQSN8J3lSrZ/YDHXHW8h5Z/XBbOHgQUCY9pxigIbAwULCQgHAgIiAgYVCgkICwIE
FgIDAQIeBwIXgAAKCRAh5Z/XBbOHgTCuA/9mYXAehM9avvq0Jm2dVbPidqxLstki
tgo3gCWmO1b5dXEBrhOZ8pZAktQ3WWoRrbwpNA7NAEIDF5l6uwMLLbGPQ5jreOdP
uzHpHONR1WWAzw2dj3v+5IcLDQ4sLi9VRgJqtMasTd8TpqMCVNcMArDBiy5hRF/e
XWEkf19Nb8qrdg==
=OEiT
-----END PGP PUBLIC KEY BLOCK-----"""


@pytest.fixture
def b_fp():
    return "8DF09DE54AB67F6031D71D6F21E59FD705B38781"


@pytest.fixture
def gpg_keys_present(gnupg, a_pubkey, b_pubkey, a_fp, b_fp):
    pubkeys = [a_pubkey, b_pubkey]
    fingerprints = [a_fp, b_fp]
    gnupg.import_keys("\n".join(pubkeys))
    present_keys = gnupg.list_keys()
    for fp in fingerprints:
        assert any(x["fingerprint"] == fp for x in present_keys)
    yield
    # cleanup is taken care of by gpghome and tmp_path


@pytest.fixture(scope="module", autouse=True)
def sig_files_present(web_root, modules):
    base = Path(RUNTIME_VARS.BASE_FILES)
    for file in [
        "custom.tar.gz",
        "custom.tar.gz.asc",
        "custom.tar.gz.SHA256",
        "custom.tar.gz.SHA256.clearsign.asc",
        "custom.tar.gz.SHA256.asc",
    ]:
        modules.file.copy(base / file, Path(web_root) / file)


@pytest.mark.skipif(HAS_GNUPG is False, reason="Needs python-gnupg library")
@pytest.mark.usefixtures("gpg_keys_present")
def test_archive_extracted_signature(tmp_path, gpghome, free_port, modules, states):
    name = tmp_path / "test_archive_extracted_signature"
    source = f"http://localhost:{free_port}/custom.tar.gz"
    signature = source + ".asc"
    source_hash = source + ".SHA256"
    ret = states.archive.extracted(
        str(name),
        source=source,
        source_hash=source_hash,
        archive_format="tar",
        options="z",
        signature=signature,
        gnupghome=str(gpghome),
    )
    assert ret.result is True
    assert ret.changes
    assert name.exists()
    assert modules.file.find(str(name))


@pytest.mark.skipif(HAS_GNUPG is False, reason="Needs python-gnupg library")
@pytest.mark.usefixtures("gpg_keys_present")
def test_archive_extracted_signature_fail(
    tmp_path, gpghome, free_port, modules, states
):
    name = tmp_path / "test_archive_extracted_signature_fail"
    source = f"http://localhost:{free_port}/custom.tar.gz"
    signature = source + ".asc"
    source_hash = source + ".SHA256"
    # although there are valid signatures, this will be denied since the one below is required
    signed_by_all = ["DEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF"]
    ret = states.archive.extracted(
        str(name),
        source=source,
        source_hash=source_hash,
        archive_format="tar",
        options="z",
        signature=signature,
        signed_by_all=signed_by_all,
        gnupghome=str(gpghome),
    )
    assert ret.result is False
    assert not ret.changes
    assert not name.exists()
    assert not modules.cp.is_cached(source)
    assert not modules.cp.is_cached(signature)


@pytest.mark.skipif(HAS_GNUPG is False, reason="Needs python-gnupg library")
@pytest.mark.usefixtures("gpg_keys_present")
@pytest.mark.parametrize("sig", [True, ".asc"])
def test_archive_extracted_source_hash_sig(
    tmp_path, sig, gpghome, free_port, modules, states
):
    name = tmp_path / "test_archive_extracted_source_hash_sig"
    source = f"http://localhost:{free_port}/custom.tar.gz"
    source_hash = source + ".SHA256"
    if sig is True:
        source_hash += ".clearsign.asc"
    else:
        sig = source_hash + sig
    ret = states.archive.extracted(
        str(name),
        source=source,
        source_hash=source_hash,
        archive_format="tar",
        options="z",
        source_hash_sig=sig,
        gnupghome=str(gpghome),
    )
    assert ret.result is True
    assert ret.changes
    assert name.exists()
    assert modules.file.find(str(name))


@pytest.mark.skipif(HAS_GNUPG is False, reason="Needs python-gnupg library")
@pytest.mark.usefixtures("gpg_keys_present")
@pytest.mark.parametrize("sig", [True, ".asc"])
def test_archive_extracted_source_hash_sig_fail(
    tmp_path, sig, gpghome, free_port, modules, states
):
    name = tmp_path / "test_archive_extracted_source_hash_sig_fail"
    source = f"http://localhost:{free_port}/custom.tar.gz"
    source_hash = source + ".SHA256.clearsign.asc"
    signed_by_any = ["DEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF"]
    ret = states.archive.extracted(
        str(name),
        source=source,
        source_hash=source_hash,
        archive_format="tar",
        options="z",
        source_hash_sig=True,
        signed_by_any=signed_by_any,
        gnupghome=str(gpghome),
    )
    assert ret.result is False
    assert not ret.changes
    assert not name.exists()
    assert not modules.cp.is_cached(source)
    assert not modules.cp.is_cached(source_hash)



"""
Tests for salt.modules.zpool

:codeauthor:    Nitin Madhok <nmadhok@g.clemson.edu>, Jorge Schrauwen <sjorge@blackdot.be>
:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils.zfs
:platform:      illumos,freebsd,linux
"""

import pytest

import salt.loader
import salt.modules.zpool as zpool
import salt.utils.decorators
import salt.utils.decorators.path
import salt.utils.zfs
from salt.utils.odict import OrderedDict
from tests.support.mock import MagicMock, patch
from tests.support.zfs import ZFSMockData

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture
def utils_patch():
    return ZFSMockData().get_patched_utils()


@pytest.fixture
def configure_loader_modules(minion_opts):
    utils = salt.loader.utils(
        minion_opts, whitelist=["zfs", "args", "systemd", "path", "platform"]
    )
    zpool_obj = {
        zpool: {
            "__opts__": minion_opts,
            "__utils__": utils,
        },
    }

    return zpool_obj


@pytest.mark.slow_test
def test_exists_success(utils_patch):
    """
    Tests successful return of exists function
    """
    ret = {}
    ret["stdout"] = (
        "NAME      SIZE  ALLOC   FREE    CAP  DEDUP  HEALTH  ALTROOT\n"
        "myzpool   149G   128K   149G     0%  1.00x  ONLINE  -"
    )
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        assert zpool.exists("myzpool")


@pytest.mark.slow_test
def test_exists_failure(utils_patch):
    """
    Tests failure return of exists function
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'myzpool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        assert not zpool.exists("myzpool")


def test_healthy(utils_patch):
    """
    Tests successful return of healthy function
    """
    ret = {}
    ret["stdout"] = "all pools are healthy"
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        assert zpool.healthy()


def test_status(utils_patch):
    """
    Tests successful return of status function
    """
    ret = {}
    ret["stdout"] = "\n".join(
        [
            "  pool: mypool",
            " state: ONLINE",
            "  scan: scrub repaired 0 in 0h6m with 0 errors on Mon Dec 21 02:06:17"
            " 2015",
            "config:",
            "",
            "\tNAME        STATE     READ WRITE CKSUM",
            "\tmypool      ONLINE       0     0     0",
            "\t  mirror-0  ONLINE       0     0     0",
            "\t    c2t0d0  ONLINE       0     0     0",
            "\t    c2t1d0  ONLINE       0     0     0",
            "",
            "errors: No known data errors",
        ]
    )
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.status()
        assert "ONLINE" == ret["mypool"]["state"]


def test_status_with_colons_in_vdevs(utils_patch):
    """
    Tests successful return of status function
    """
    ret = {}
    ret["stdout"] = "\n".join(
        [
            "  pool: mypool",
            " state: ONLINE",
            "  scan: scrub repaired 0 in 0h6m with 0 errors on Mon Dec 21 02:06:17"
            " 2015",
            "config:",
            "",
            "\tNAME        STATE     READ WRITE CKSUM",
            "\tmypool      ONLINE       0     0     0",
            "\t  mirror-0  ONLINE       0     0     0",
            "\t    usb-WD_My_Book_Duo_25F6_....32-0:0  ONLINE       0     0     0",
            "\t    usb-WD_My_Book_Duo_25F6_....32-0:1  ONLINE       0     0     0",
            "",
            "errors: No known data errors",
        ]
    )
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.status()
        assert "ONLINE" == ret["mypool"]["state"]


@pytest.mark.slow_test
def test_iostat(utils_patch):
    """
    Tests successful return of iostat function
    """
    ret = {}
    ret["stdout"] = "\n".join(
        [
            "               capacity     operations    bandwidth",
            "pool        alloc   free   read  write   read  write",
            "----------  -----  -----  -----  -----  -----  -----",
            "mypool      46.7G  64.3G      4     19   113K   331K",
            "  mirror    46.7G  64.3G      4     19   113K   331K",
            "    c2t0d0      -      -      1     10   114K   334K",
            "    c2t1d0      -      -      1     10   114K   334K",
            "----------  -----  -----  -----  -----  -----  -----",
        ]
    )
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.iostat("mypool", parsable=False)
        assert "46.7G" == ret["mypool"]["capacity-alloc"]


def test_iostat_parsable(utils_patch):
    """
    Tests successful return of iostat function

    .. note:
        The command output is the same as the non parsable!
        There is no -p flag for zpool iostat, but our type
        conversions can handle this!
    """
    ret = {}
    ret["stdout"] = "\n".join(
        [
            "               capacity     operations    bandwidth",
            "pool        alloc   free   read  write   read  write",
            "----------  -----  -----  -----  -----  -----  -----",
            "mypool      46.7G  64.3G      4     19   113K   331K",
            "  mirror    46.7G  64.3G      4     19   113K   331K",
            "    c2t0d0      -      -      1     10   114K   334K",
            "    c2t1d0      -      -      1     10   114K   334K",
            "----------  -----  -----  -----  -----  -----  -----",
        ]
    )
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.iostat("mypool", parsable=True)
        assert 50143743180 == ret["mypool"]["capacity-alloc"]


def test_list(utils_patch):
    """
    Tests successful return of list function
    """
    ret = {}
    ret["stdout"] = "mypool\t1.81T\t661G\t1.17T\t35%\t11%\tONLINE"
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.list_(parsable=False)
        res = OrderedDict(
            [
                (
                    "mypool",
                    OrderedDict(
                        [
                            ("size", "1.81T"),
                            ("alloc", "661G"),
                            ("free", "1.17T"),
                            ("cap", "35%"),
                            ("frag", "11%"),
                            ("health", "ONLINE"),
                        ]
                    ),
                )
            ]
        )
        assert ret == res


@pytest.mark.slow_test
def test_list_parsable(utils_patch):
    """
    Tests successful return of list function with parsable output
    """
    ret = {}
    ret["stdout"] = "mypool\t1.81T\t661G\t1.17T\t35%\t11%\tONLINE"
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.list_(parsable=True)
        res = OrderedDict(
            [
                (
                    "mypool",
                    OrderedDict(
                        [
                            ("size", 1990116046274),
                            ("alloc", 709743345664),
                            ("free", 1286428604497),
                            ("cap", "35%"),
                            ("frag", "11%"),
                            ("health", "ONLINE"),
                        ]
                    ),
                )
            ]
        )
        assert ret == res


def test_get(utils_patch):
    """
    Tests successful return of get function
    """
    ret = {}
    ret["stdout"] = "mypool\tsize\t1.81T\t-\n"
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.get("mypool", "size", parsable=False)
        res = OrderedDict(OrderedDict([("size", "1.81T")]))
        assert ret == res


@pytest.mark.slow_test
def test_get_parsable(utils_patch):
    """
    Tests successful return of get function with parsable output
    """
    ret = {}
    ret["stdout"] = "mypool\tsize\t1.81T\t-\n"
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.get("mypool", "size", parsable=True)
        res = OrderedDict(OrderedDict([("size", 1990116046274)]))
        assert ret == res


@pytest.mark.slow_test
def test_get_whitespace(utils_patch):
    """
    Tests successful return of get function with a string with whitespaces
    """
    ret = {}
    ret["stdout"] = "mypool\tcomment\tmy testing pool\t-\n"
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.get("mypool", "comment")
        res = OrderedDict(OrderedDict([("comment", "my testing pool")]))
        assert ret == res


@pytest.mark.slow_test
def test_scrub_start(utils_patch):
    """
    Tests start of scrub
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    mock_exists = MagicMock(return_value=True)

    with patch.dict(zpool.__salt__, {"zpool.exists": mock_exists}), patch.dict(
        zpool.__salt__, {"cmd.run_all": mock_cmd}
    ), patch.dict(zpool.__utils__, utils_patch):
        ret = zpool.scrub("mypool")
        res = OrderedDict(OrderedDict([("scrubbing", True)]))
        assert ret == res


@pytest.mark.slow_test
def test_scrub_pause(utils_patch):
    """
    Tests pause of scrub
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    mock_exists = MagicMock(return_value=True)

    with patch.dict(zpool.__salt__, {"zpool.exists": mock_exists}), patch.dict(
        zpool.__salt__, {"cmd.run_all": mock_cmd}
    ), patch.dict(zpool.__utils__, utils_patch):
        ret = zpool.scrub("mypool", pause=True)
        res = OrderedDict(OrderedDict([("scrubbing", False)]))
        assert ret == res


@pytest.mark.slow_test
def test_scrub_stop(utils_patch):
    """
    Tests pauze of scrub
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    mock_exists = MagicMock(return_value=True)

    with patch.dict(zpool.__salt__, {"zpool.exists": mock_exists}), patch.dict(
        zpool.__salt__, {"cmd.run_all": mock_cmd}
    ), patch.dict(zpool.__utils__, utils_patch):
        ret = zpool.scrub("mypool", stop=True)
        res = OrderedDict(OrderedDict([("scrubbing", False)]))
        assert ret == res


def test_split_success(utils_patch):
    """
    Tests split on success
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.split("datapool", "backuppool")
        res = OrderedDict([("split", True)])
        assert ret == res


@pytest.mark.slow_test
def test_split_exist_new(utils_patch):
    """
    Tests split on exising new pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "Unable to split datapool: pool already exists"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.split("datapool", "backuppool")
        res = OrderedDict(
            [
                ("split", False),
                ("error", "Unable to split datapool: pool already exists"),
            ]
        )
        assert ret == res


def test_split_missing_pool(utils_patch):
    """
    Tests split on missing source pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'datapool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.split("datapool", "backuppool")
        res = OrderedDict(
            [("split", False), ("error", "cannot open 'datapool': no such pool")]
        )
        assert ret == res


@pytest.mark.slow_test
def test_split_not_mirror(utils_patch):
    """
    Tests split on source pool is not a mirror
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = (
        "Unable to split datapool: Source pool must be composed only of mirrors"
    )
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.split("datapool", "backuppool")
        res = OrderedDict(
            [
                ("split", False),
                (
                    "error",
                    "Unable to split datapool: Source pool must be composed only of"
                    " mirrors",
                ),
            ]
        )
        assert ret == res


def test_labelclear_success(utils_patch):
    """
    Tests labelclear on successful label removal
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.labelclear("/dev/rdsk/c0t0d0", force=False)
        res = OrderedDict([("labelcleared", True)])
        assert ret == res


def test_labelclear_nodevice(utils_patch):
    """
    Tests labelclear on non existing device
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "failed to open /dev/rdsk/c0t0d0: No such file or directory"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.labelclear("/dev/rdsk/c0t0d0", force=False)
        res = OrderedDict(
            [
                ("labelcleared", False),
                (
                    "error",
                    "failed to open /dev/rdsk/c0t0d0: No such file or directory",
                ),
            ]
        )
        assert ret == res


def test_labelclear_cleared(utils_patch):
    """
    Tests labelclear on device with no label
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "failed to read label from /dev/rdsk/c0t0d0"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.labelclear("/dev/rdsk/c0t0d0", force=False)
        res = OrderedDict(
            [
                ("labelcleared", False),
                ("error", "failed to read label from /dev/rdsk/c0t0d0"),
            ]
        )
        assert ret == res


def test_labelclear_exported(utils_patch):
    """
    Tests labelclear on device with from exported pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "\n".join(
        [
            "use '-f' to override the following error:",
            '/dev/rdsk/c0t0d0 is a member of exported pool "mypool"',
        ]
    )
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.labelclear("/dev/rdsk/c0t0d0", force=False)
        res = OrderedDict(
            [
                ("labelcleared", False),
                (
                    "error",
                    "use 'force=True' to override the following"
                    " error:\n/dev/rdsk/c0t0d0 is a member of exported pool"
                    ' "mypool"',
                ),
            ]
        )
        assert ret == res


@pytest.mark.skip_if_binaries_missing("mkfile", reason="Cannot find mkfile executable")
def test_create_file_vdev_success(utils_patch):
    """
    Tests create_file_vdev when out of space
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.create_file_vdev("64M", "/vdisks/disk0")
        res = OrderedDict([("/vdisks/disk0", "created")])
        assert ret == res


@pytest.mark.skip_if_binaries_missing("mkfile", reason="Cannot find mkfile executable")
def test_create_file_vdev_nospace(utils_patch):
    """
    Tests create_file_vdev when out of space
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = (
        "/vdisks/disk0: initialized 10424320 of 67108864 bytes: No space left on"
        " device"
    )
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.create_file_vdev("64M", "/vdisks/disk0")
        res = OrderedDict(
            [
                ("/vdisks/disk0", "failed"),
                (
                    "error",
                    OrderedDict(
                        [
                            (
                                "/vdisks/disk0",
                                " initialized 10424320 of 67108864 bytes: No space"
                                " left on device",
                            ),
                        ]
                    ),
                ),
            ]
        )
        assert ret == res


def test_export_success(utils_patch):
    """
    Tests export
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.export("mypool")
        res = OrderedDict([("exported", True)])
        assert ret == res


@pytest.mark.slow_test
def test_export_nopool(utils_patch):
    """
    Tests export when the pool does not exists
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'mypool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.export("mypool")
        res = OrderedDict(
            [("exported", False), ("error", "cannot open 'mypool': no such pool")]
        )
        assert ret == res


@pytest.mark.slow_test
def test_import_success(utils_patch):
    """
    Tests import
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.import_("mypool")
        res = OrderedDict([("imported", True)])
        assert ret == res


def test_import_duplicate(utils_patch):
    """
    Tests import with already imported pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "\n".join(
        [
            "cannot import 'mypool': a pool with that name already exists",
            "use the form 'zpool import <pool | id> <newpool>' to give it a new"
            " name",
        ]
    )
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.import_("mypool")
        res = OrderedDict(
            [
                ("imported", False),
                (
                    "error",
                    "cannot import 'mypool': a pool with that name already"
                    " exists\nuse the form 'zpool import <pool | id> <newpool>' to"
                    " give it a new name",
                ),
            ]
        )
        assert ret == res


def test_import_nopool(utils_patch):
    """
    Tests import
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot import 'mypool': no such pool available"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.import_("mypool")
        res = OrderedDict(
            [
                ("imported", False),
                ("error", "cannot import 'mypool': no such pool available"),
            ]
        )
        assert ret == res


@pytest.mark.slow_test
def test_online_success(utils_patch):
    """
    Tests online
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.online("mypool", "/dev/rdsk/c0t0d0")
        res = OrderedDict([("onlined", True)])
        assert ret == res


def test_online_nodevice(utils_patch):
    """
    Tests online
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot online /dev/rdsk/c0t0d1: no such device in pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.online("mypool", "/dev/rdsk/c0t0d1")
        res = OrderedDict(
            [
                ("onlined", False),
                ("error", "cannot online /dev/rdsk/c0t0d1: no such device in pool"),
            ]
        )
        assert ret == res


def test_offline_success(utils_patch):
    """
    Tests offline
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.offline("mypool", "/dev/rdsk/c0t0d0")
        res = OrderedDict([("offlined", True)])
        assert ret == res


def test_offline_nodevice(utils_patch):
    """
    Tests offline
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot offline /dev/rdsk/c0t0d1: no such device in pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.offline("mypool", "/dev/rdsk/c0t0d1")
        res = OrderedDict(
            [
                ("offlined", False),
                (
                    "error",
                    "cannot offline /dev/rdsk/c0t0d1: no such device in pool",
                ),
            ]
        )
        assert ret == res


def test_offline_noreplica(utils_patch):
    """
    Tests offline
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot offline /dev/rdsk/c0t0d1: no valid replicas"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.offline("mypool", "/dev/rdsk/c0t0d1")
        res = OrderedDict(
            [
                ("offlined", False),
                ("error", "cannot offline /dev/rdsk/c0t0d1: no valid replicas"),
            ]
        )
        assert ret == res


@pytest.mark.slow_test
def test_reguid_success(utils_patch):
    """
    Tests reguid
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.reguid("mypool")
        res = OrderedDict([("reguided", True)])
        assert ret == res


@pytest.mark.slow_test
def test_reguid_nopool(utils_patch):
    """
    Tests reguid with missing pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'mypool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.reguid("mypool")
        res = OrderedDict(
            [("reguided", False), ("error", "cannot open 'mypool': no such pool")]
        )
        assert ret == res


@pytest.mark.slow_test
def test_reopen_success(utils_patch):
    """
    Tests reopen
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.reopen("mypool")
        res = OrderedDict([("reopened", True)])
        assert ret == res


def test_reopen_nopool(utils_patch):
    """
    Tests reopen with missing pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'mypool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.reopen("mypool")
        res = OrderedDict(
            [("reopened", False), ("error", "cannot open 'mypool': no such pool")]
        )
        assert ret == res


def test_upgrade_success(utils_patch):
    """
    Tests upgrade
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.upgrade("mypool")
        res = OrderedDict([("upgraded", True)])
        assert ret == res


def test_upgrade_nopool(utils_patch):
    """
    Tests upgrade with missing pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'mypool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.upgrade("mypool")
        res = OrderedDict(
            [("upgraded", False), ("error", "cannot open 'mypool': no such pool")]
        )
        assert ret == res


@pytest.mark.slow_test
def test_history_success(utils_patch):
    """
    Tests history
    """
    ret = {}
    ret["stdout"] = "\n".join(
        [
            "History for 'mypool':",
            "2018-01-18.16:56:12 zpool create -f mypool /dev/rdsk/c0t0d0",
            "2018-01-19.16:01:55 zpool attach -f mypool /dev/rdsk/c0t0d0"
            " /dev/rdsk/c0t0d1",
        ]
    )
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.history("mypool")
        res = OrderedDict(
            [
                (
                    "mypool",
                    OrderedDict(
                        [
                            (
                                "2018-01-18.16:56:12",
                                "zpool create -f mypool /dev/rdsk/c0t0d0",
                            ),
                            (
                                "2018-01-19.16:01:55",
                                "zpool attach -f mypool /dev/rdsk/c0t0d0"
                                " /dev/rdsk/c0t0d1",
                            ),
                        ]
                    ),
                ),
            ]
        )
        assert ret == res


def test_history_nopool(utils_patch):
    """
    Tests history with missing pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'mypool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.history("mypool")
        res = OrderedDict([("error", "cannot open 'mypool': no such pool")])
        assert ret == res


def test_clear_success(utils_patch):
    """
    Tests clear
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.clear("mypool")
        res = OrderedDict([("cleared", True)])
        assert ret == res


def test_clear_nopool(utils_patch):
    """
    Tests clear with missing pool
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot open 'mypool': no such pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.clear("mypool")
        res = OrderedDict(
            [("cleared", False), ("error", "cannot open 'mypool': no such pool")]
        )


def test_clear_nodevice(utils_patch):
    """
    Tests clear with non existign device
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "cannot clear errors for /dev/rdsk/c0t0d0: no such device in pool"
    ret["retcode"] = 1
    mock_cmd = MagicMock(return_value=ret)

    with patch.dict(zpool.__salt__, {"cmd.run_all": mock_cmd}), patch.dict(
        zpool.__utils__, utils_patch
    ):
        ret = zpool.clear("mypool", "/dev/rdsk/c0t0d0")
        res = OrderedDict(
            [
                ("cleared", False),
                (
                    "error",
                    "cannot clear errors for /dev/rdsk/c0t0d0: no such device in"
                    " pool",
                ),
            ]
        )
        assert ret == res



import pytest

import salt.config
import salt.loader
import salt.roster.sshknownhosts as sshknownhosts
from tests.support.mock import patch


@pytest.fixture
def ALL_DICT():
    yield {
        "server1": {"host": "server1"},
        "server2": {"host": "server2"},
        "server3.local": {"host": "server3.local"},
        "eu-mysql-1.local": {"host": "eu-mysql-1.local"},
        "eu-mysql-2": {"host": "eu-mysql-2"},
        "eu-mysql-2.local": {"host": "eu-mysql-2.local"},
    }


@pytest.fixture
def GLOB_DICT():
    yield {
        "server1": {"host": "server1"},
        "server2": {"host": "server2"},
        "server3.local": {"host": "server3.local"},
    }


@pytest.fixture
def PCRE_DICT():
    yield {"eu-mysql-2": {"host": "eu-mysql-2"}}


@pytest.fixture
def known_hosts():

    yield """
server1 ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBD0vTE0R76xiKEXAdebZW0a3xGLeP2Fet/5YHQgprry3wuXzjBJwGcm8PVFNfbK/C7oAgFUg8NVX7xqQnScekJg=
server2 ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAQEAui+dKujjKF92dDdM9hZzCd+BdTDRnvsWqMf88NjushOmFCt/8zXbB1TvYQmdCcy1qXqhmkbgUdtVuLHnhncf/niCtyih3K3ZR7NpecBydcC+0xv0UeXk/xCGcwM2V0BuukrV/5qRqhyG0rK1hd+Iv9fkB0/s8D/HLcEB1/V4g77XxPGnI7lNANFbZpWs1LrnAec7JIkHO9MHEfuhQWZR6+/iIXIwQoc1RCToQbWQFCYFwrnDrAUHC2+izJiP2VDNW6xboVcf6DpwydfYvFdM8Mo97DEcchlwIWhmGl//LpnwafujFZCE5vDveA8X4uKZEXxoCmUPIGfkx6xIzzTkqQ==
server3.local ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBAgKWpCT7JIeK/qzwE5lUQkLRfkRa5WnyyeF+aYCKDUHB4b4Pn+acm8FOca+riulPDY/gJhb0MX3Rf/t6MrEHQA=
eu-mysql-1.local ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAQEAsuToIp6iqJ3lHPQzCTiNf5F8uf/CjAljuxRjURYCbQydts2lnnqTpjamL1b8/FpvB1dDlA71G79yTftVZ8EqL2VaN0tL242MXaqy2nmeVjy89dtOyk35IHwQe8Bi6mu3vLYCFnysiAvrtLQMFe8jNjndsvf27LNKox8pIAyOyN3hONL+bXEcPB2RjIUL8wS8uTeOueuPbVwc1cHkUuMjlNzsH3l6KMVjJZ8keFdRj8iogV8oZGR3KGoPfX4aZDt9S+L/k97fWkOhSKLWkKbplEcmIjuF5pgZLO3Wf35eLZN12PcHuX7WFWZi+UxjJDW2VLaP867La4YXDEU3LNdPEQ==
eu-mysql-2,eu-mysql-2.local ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzdHAyNTYAAABBBHFnjWT+gnUGRA2zW+LGZdebSkUVKBb6F/XCcDrtBZmaxCNaS/+F6SYzXP4MghCQhXFEPd7MpFnwPV8giU1NUag="""


@pytest.fixture
def configure_loader_modules(salt_master_factory, tmp_path):
    opts = salt_master_factory.config.copy()
    utils = salt.loader.utils(opts)
    runner = salt.loader.runner(opts, utils=utils)

    return {
        sshknownhosts: {
            "__opts__": {},
            "__runner__": runner,
            "__utils__": utils,
        }
    }


def test_all(known_hosts, tmp_path, ALL_DICT):
    with pytest.helpers.temp_file(
        "known_hosts", known_hosts, directory=tmp_path
    ) as known_hosts_file:
        opts = {"ssh_known_hosts_file": str(known_hosts_file)}
        with patch.dict(sshknownhosts.__opts__, opts):
            targets = sshknownhosts.targets(tgt="*")
            assert targets == ALL_DICT


def test_glob(known_hosts, tmp_path, GLOB_DICT):
    with pytest.helpers.temp_file(
        "known_hosts", known_hosts, directory=tmp_path
    ) as known_hosts_file:
        opts = {"ssh_known_hosts_file": str(known_hosts_file)}
        with patch.dict(sshknownhosts.__opts__, opts):
            targets = sshknownhosts.targets(tgt="server*")
            assert targets == GLOB_DICT


def test_pcre(known_hosts, tmp_path, PCRE_DICT):
    with pytest.helpers.temp_file(
        "known_hosts", known_hosts, directory=tmp_path
    ) as known_hosts_file:
        opts = {"ssh_known_hosts_file": str(known_hosts_file)}
        with patch.dict(sshknownhosts.__opts__, opts):
            targets = sshknownhosts.targets(tgt="eu-mysql-2$", tgt_type="pcre")
            assert targets == PCRE_DICT



"""
Test the custom grains
"""

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.slow_test,
]


def test_grains_passed_to_custom_grain(salt_call_cli):
    """
    test if current grains are passed to grains module functions that have a grains argument
    """
    ret = salt_call_cli.run("grains.item", "custom_grain_test")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["custom_grain_test"] == "itworked"



import salt.utils.parsers
from salt.exceptions import SaltInvocationError
from salt.utils.verify import check_user


class SaltKey(salt.utils.parsers.SaltKeyOptionParser):
    """
    Initialize the Salt key manager
    """

    def run(self):
        """
        Execute salt-key
        """
        import salt.key

        self.parse_args()
        if self.options.delete_all:
            if self.args:
                raise SaltInvocationError(
                    "Delete all takes no arguments. Use -d to delete specified keys"
                )

        key = salt.key.KeyCLI(self.config)
        if check_user(self.config["user"]):
            key.run()



"""
Execution module for Amazon Elasticache using boto3
===================================================

.. versionadded:: 2017.7.0

:configuration: This module accepts explicit elasticache credentials but can
    also utilize IAM roles assigned to the instance through Instance Profiles.
    Dynamic credentials are then automatically obtained from AWS API and no
    further configuration is necessary. More Information available at:

    .. code-block:: text

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html

    If IAM roles are not used you need to specify them either in a pillar or
    in the minion's config file:

    .. code-block:: yaml

        elasticache.keyid: GKTADJGHEIQSXMKKRBJ08H
        elasticache.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    A region may also be specified in the configuration:

    .. code-block:: yaml

        elasticache.region: us-east-1

    If a region is not specified, the default is us-east-1.

    It's also possible to specify key, keyid and region via a profile, either
    as a passed in dict, or as a string to pull from pillars or minion config:

    .. code-block:: yaml

        myprofile:
            keyid: GKTADJGHEIQSXMKKRBJ08H
            key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

:depends: boto3
"""

# keep lint from choking on _get_conn and _cache_id
# pylint: disable=E0602


import logging
import time

import salt.utils.compat
import salt.utils.versions
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

try:
    # pylint: disable=unused-import
    import boto3
    import botocore

    # pylint: enable=unused-import
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False


def __virtual__():
    """
    Only load if boto libraries exist and if boto libraries are greater than
    a given version.
    """
    return salt.utils.versions.check_boto_reqs()


def __init__(opts):
    if HAS_BOTO3:
        __utils__["boto3.assign_funcs"](
            __name__,
            "elasticache",
            get_conn_funcname="_get_conn",
            cache_id_funcname="_cache_id",
            exactly_one_funcname=None,
        )


def _collect_results(func, item, args, marker="Marker"):
    ret = []
    Marker = args[marker] if marker in args else ""
    while Marker is not None:
        r = func(**args)
        ret += r.get(item)
        Marker = r.get(marker)
        args.update({marker: Marker})
    return ret


def _describe_resource(
    name=None,
    name_param=None,
    res_type=None,
    info_node=None,
    conn=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    if conn is None:
        conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    try:
        func = "describe_" + res_type + "s"
        f = getattr(conn, func)
    except (AttributeError, KeyError) as e:
        raise SaltInvocationError(f"No function '{func}()' found: {e.message}")
    # Undocumented, but you can't pass 'Marker' if searching for a specific resource...
    args.update({name_param: name} if name else {"Marker": ""})
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        return _collect_results(f, info_node, args)
    except botocore.exceptions.ClientError as e:
        log.debug(e)
        return None


def _delete_resource(
    name,
    name_param,
    desc,
    res_type,
    wait=0,
    status_param=None,
    status_gone="deleted",
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    """
    Delete a generic Elasticache resource.
    """
    try:
        wait = int(wait)
    except Exception:  # pylint: disable=broad-except
        raise SaltInvocationError(
            "Bad value ('{}') passed for 'wait' param - must be an "
            "int or boolean.".format(wait)
        )
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if name_param in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided '%s: %s'",
            name,
            name_param,
            args[name_param],
        )
        name = args[name_param]
    else:
        args[name_param] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        func = "delete_" + res_type
        f = getattr(conn, func)
        if wait:
            func = "describe_" + res_type + "s"
            s = globals()[func]
    except (AttributeError, KeyError) as e:
        raise SaltInvocationError(f"No function '{func}()' found: {e.message}")
    try:

        f(**args)
        if not wait:
            log.info("%s %s deletion requested.", desc.title(), name)
            return True
        log.info("Waiting up to %s seconds for %s %s to be deleted.", wait, desc, name)
        orig_wait = wait
        while wait > 0:
            r = s(name=name, conn=conn)
            if not r or r[0].get(status_param) == status_gone:
                log.info("%s %s deleted.", desc.title(), name)
                return True
            sleep = wait if wait % 60 == wait else 60
            log.info("Sleeping %s seconds for %s %s to be deleted.", sleep, desc, name)
            time.sleep(sleep)
            wait -= sleep
        log.error("%s %s not deleted after %s seconds!", desc.title(), name, orig_wait)

        return False
    except botocore.exceptions.ClientError as e:
        log.error("Failed to delete %s %s: %s", desc, name, e)
        return False


def _create_resource(
    name,
    name_param=None,
    desc=None,
    res_type=None,
    wait=0,
    status_param=None,
    status_good="available",
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    try:
        wait = int(wait)
    except Exception:  # pylint: disable=broad-except
        raise SaltInvocationError(
            "Bad value ('{}') passed for 'wait' param - must be an "
            "int or boolean.".format(wait)
        )
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if name_param in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided '%s: %s'",
            name,
            name_param,
            args[name_param],
        )
        name = args[name_param]
    else:
        args[name_param] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        func = "create_" + res_type
        f = getattr(conn, func)
        if wait:
            func = "describe_" + res_type + "s"
            s = globals()[func]
    except (AttributeError, KeyError) as e:
        raise SaltInvocationError(f"No function '{func}()' found: {e.message}")
    try:
        f(**args)
        if not wait:
            log.info("%s %s created.", desc.title(), name)
            return True
        log.info(
            "Waiting up to %s seconds for %s %s to be become available.",
            wait,
            desc,
            name,
        )
        orig_wait = wait
        while wait > 0:
            r = s(name=name, conn=conn)
            if r and r[0].get(status_param) == status_good:
                log.info("%s %s created and available.", desc.title(), name)
                return True
            sleep = wait if wait % 60 == wait else 60
            log.info(
                "Sleeping %s seconds for %s %s to become available.", sleep, desc, name
            )
            time.sleep(sleep)
            wait -= sleep
        log.error(
            "%s %s not available after %s seconds!", desc.title(), name, orig_wait
        )
        return False
    except botocore.exceptions.ClientError as e:
        msg = f"Failed to create {desc} {name}: {e}"
        log.error(msg)
        return False


def _modify_resource(
    name,
    name_param=None,
    desc=None,
    res_type=None,
    wait=0,
    status_param=None,
    status_good="available",
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    try:
        wait = int(wait)
    except Exception:  # pylint: disable=broad-except
        raise SaltInvocationError(
            "Bad value ('{}') passed for 'wait' param - must be an "
            "int or boolean.".format(wait)
        )
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if name_param in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided '%s: %s'",
            name,
            name_param,
            args[name_param],
        )
        name = args[name_param]
    else:
        args[name_param] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        func = "modify_" + res_type
        f = getattr(conn, func)
        if wait:
            func = "describe_" + res_type + "s"
            s = globals()[func]
    except (AttributeError, KeyError) as e:
        raise SaltInvocationError(f"No function '{func}()' found: {e.message}")
    try:
        f(**args)
        if not wait:
            log.info("%s %s modification requested.", desc.title(), name)
            return True
        log.info(
            "Waiting up to %s seconds for %s %s to be become available.",
            wait,
            desc,
            name,
        )
        orig_wait = wait
        while wait > 0:
            r = s(name=name, conn=conn)
            if r and r[0].get(status_param) == status_good:
                log.info("%s %s modified and available.", desc.title(), name)
                return True
            sleep = wait if wait % 60 == wait else 60
            log.info(
                "Sleeping %s seconds for %s %s to become available.", sleep, desc, name
            )
            time.sleep(sleep)
            wait -= sleep
        log.error(
            "%s %s not available after %s seconds!", desc.title(), name, orig_wait
        )
        return False
    except botocore.exceptions.ClientError as e:
        msg = f"Failed to modify {desc} {name}: {e}"
        log.error(msg)
        return False


def describe_cache_clusters(
    name=None, conn=None, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Return details about all (or just one) Elasticache cache clusters.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.describe_cache_clusters
        salt myminion boto3_elasticache.describe_cache_clusters myelasticache
    """
    return _describe_resource(
        name=name,
        name_param="CacheClusterId",
        res_type="cache_cluster",
        info_node="CacheClusters",
        conn=conn,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def cache_cluster_exists(
    name, conn=None, region=None, key=None, keyid=None, profile=None
):
    """
    Check to see if a cache cluster exists.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.cache_cluster_exists myelasticache
    """
    return bool(
        describe_cache_clusters(
            name=name, conn=conn, region=region, key=key, keyid=keyid, profile=profile
        )
    )


def create_cache_cluster(
    name,
    wait=600,
    security_groups=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    """
    Create a cache cluster.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.create_cache_cluster name=myCacheCluster \
                                                             Engine=redis \
                                                             CacheNodeType=cache.t2.micro \
                                                             NumCacheNodes=1 \
                                                             SecurityGroupIds='[sg-11223344]' \
                                                             CacheSubnetGroupName=myCacheSubnetGroup
    """
    if security_groups:
        if not isinstance(security_groups, list):
            security_groups = [security_groups]
        sgs = __salt__["boto_secgroup.convert_to_group_ids"](
            groups=security_groups, region=region, key=key, keyid=keyid, profile=profile
        )
        if "SecurityGroupIds" not in args:
            args["SecurityGroupIds"] = []
        args["SecurityGroupIds"] += sgs
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    return _create_resource(
        name,
        name_param="CacheClusterId",
        desc="cache cluster",
        res_type="cache_cluster",
        wait=wait,
        status_param="CacheClusterStatus",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def modify_cache_cluster(
    name,
    wait=600,
    security_groups=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    """
    Update a cache cluster in place.

    Notes:  {ApplyImmediately: False} is pretty danged silly in the context of salt.
            You can pass it, but for fairly obvious reasons the results over multiple
            runs will be undefined and probably contrary to your desired state.
            Reducing the number of nodes requires an EXPLICIT CacheNodeIdsToRemove be
            passed, which until a reasonable heuristic for programmatically deciding
            which nodes to remove has been established, MUST be decided and populated
            intentionally before a state call, and removed again before the next.  In
            practice this is not particularly useful and should probably be avoided.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.create_cache_cluster name=myCacheCluster \
                                                             NotificationTopicStatus=inactive
    """
    if security_groups:
        if not isinstance(security_groups, list):
            security_groups = [security_groups]
        sgs = __salt__["boto_secgroup.convert_to_group_ids"](
            groups=security_groups, region=region, key=key, keyid=keyid, profile=profile
        )
        if "SecurityGroupIds" not in args:
            args["SecurityGroupIds"] = []
        args["SecurityGroupIds"] += sgs
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    return _modify_resource(
        name,
        name_param="CacheClusterId",
        desc="cache cluster",
        res_type="cache_cluster",
        wait=wait,
        status_param="CacheClusterStatus",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def delete_cache_cluster(
    name, wait=600, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Delete a cache cluster.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.delete myelasticache
    """
    return _delete_resource(
        name,
        name_param="CacheClusterId",
        desc="cache cluster",
        res_type="cache_cluster",
        wait=wait,
        status_param="CacheClusterStatus",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def describe_replication_groups(
    name=None, conn=None, region=None, key=None, keyid=None, profile=None
):
    """
    Return details about all (or just one) Elasticache replication groups.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.describe_replication_groups
        salt myminion boto3_elasticache.describe_replication_groups myelasticache
    """
    return _describe_resource(
        name=name,
        name_param="ReplicationGroupId",
        res_type="replication_group",
        info_node="ReplicationGroups",
        conn=conn,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )


def replication_group_exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if a replication group exists.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.replication_group_exists myelasticache
    """
    return bool(
        describe_replication_groups(
            name=name, region=region, key=key, keyid=keyid, profile=profile
        )
    )


def create_replication_group(
    name,
    wait=600,
    security_groups=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    """
    Create a replication group.
    Params are extensive and variable - see
    http://boto3.readthedocs.io/en/latest/reference/services/elasticache.html?#ElastiCache.Client.create_replication_group
    for in-depth usage documentation.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.create_replication_group \
                                                  name=myelasticache \
                                                  ReplicationGroupDescription=description
    """
    if security_groups:
        if not isinstance(security_groups, list):
            security_groups = [security_groups]
        sgs = __salt__["boto_secgroup.convert_to_group_ids"](
            groups=security_groups, region=region, key=key, keyid=keyid, profile=profile
        )
        if "SecurityGroupIds" not in args:
            args["SecurityGroupIds"] = []
        args["SecurityGroupIds"] += sgs
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    return _create_resource(
        name,
        name_param="ReplicationGroupId",
        desc="replication group",
        res_type="replication_group",
        wait=wait,
        status_param="Status",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def modify_replication_group(
    name,
    wait=600,
    security_groups=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    **args,
):
    """
    Modify a replication group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.modify_replication_group \
                                                  name=myelasticache \
                                                  ReplicationGroupDescription=newDescription
    """
    if security_groups:
        if not isinstance(security_groups, list):
            security_groups = [security_groups]
        sgs = __salt__["boto_secgroup.convert_to_group_ids"](
            groups=security_groups, region=region, key=key, keyid=keyid, profile=profile
        )
        if "SecurityGroupIds" not in args:
            args["SecurityGroupIds"] = []
        args["SecurityGroupIds"] += sgs
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    return _modify_resource(
        name,
        name_param="ReplicationGroupId",
        desc="replication group",
        res_type="replication_group",
        wait=wait,
        status_param="Status",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def delete_replication_group(
    name, wait=600, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Delete an ElastiCache replication group, optionally taking a snapshot first.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.delete_replication_group my-replication-group
    """
    return _delete_resource(
        name,
        name_param="ReplicationGroupId",
        desc="replication group",
        res_type="replication_group",
        wait=wait,
        status_param="Status",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def describe_cache_subnet_groups(
    name=None, conn=None, region=None, key=None, keyid=None, profile=None
):
    """
    Return details about all (or just one) Elasticache replication groups.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.describe_cache_subnet_groups region=us-east-1
    """
    return _describe_resource(
        name=name,
        name_param="CacheSubnetGroupName",
        res_type="cache_subnet_group",
        info_node="CacheSubnetGroups",
        conn=conn,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )


def cache_subnet_group_exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if an ElastiCache subnet group exists.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.cache_subnet_group_exists my-subnet-group
    """
    return bool(
        describe_cache_subnet_groups(
            name=name, region=region, key=key, keyid=keyid, profile=profile
        )
    )


def list_cache_subnet_groups(region=None, key=None, keyid=None, profile=None):
    """
    Return a list of all cache subnet group names

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.list_cache_subnet_groups region=us-east-1
    """
    return [
        g["CacheSubnetGroupName"]
        for g in describe_cache_subnet_groups(None, region, key, keyid, profile)
    ]


def create_cache_subnet_group(
    name, subnets=None, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Create an ElastiCache subnet group

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.create_cache_subnet_group name=my-subnet-group \
                                              CacheSubnetGroupDescription="description" \
                                              subnets='[myVPCSubnet1,myVPCSubnet2]'
    """
    if subnets:
        if "SubnetIds" not in args:
            args["SubnetIds"] = []
        if not isinstance(subnets, list):
            subnets = [subnets]
        for subnet in subnets:
            if subnet.startswith("subnet-"):
                # Moderately safe assumption... :)  Will be caught further down if incorrect.
                args["SubnetIds"] += [subnet]
                continue
            sn = __salt__["boto_vpc.describe_subnets"](
                subnet_names=subnet,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            ).get("subnets")
            if not sn:
                raise SaltInvocationError(
                    f"Could not resolve Subnet Name {subnet} to an ID."
                )
            if len(sn) == 1:
                args["SubnetIds"] += [sn[0]["id"]]
            elif len(sn) > 1:
                raise CommandExecutionError(
                    f"Subnet Name {subnet} returned more than one ID."
                )
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    return _create_resource(
        name,
        name_param="CacheSubnetGroupName",
        desc="cache subnet group",
        res_type="cache_subnet_group",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def modify_cache_subnet_group(
    name, subnets=None, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Modify an ElastiCache subnet group

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.modify_cache_subnet_group \
                                              name=my-subnet-group \
                                              subnets='[myVPCSubnet3]'
    """
    if subnets:
        if "SubnetIds" not in args:
            args["SubnetIds"] = []
        if not isinstance(subnets, list):
            subnets = [subnets]
        for subnet in subnets:
            sn = __salt__["boto_vpc.describe_subnets"](
                subnet_names=subnet,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            ).get("subnets")
            if len(sn) == 1:
                args["SubnetIds"] += [sn[0]["id"]]
            elif len(sn) > 1:
                raise CommandExecutionError(
                    f"Subnet Name {subnet} returned more than one ID."
                )
            elif subnet.startswith("subnet-"):
                # Moderately safe assumption... :)  Will be caught later if incorrect.
                args["SubnetIds"] += [subnet]
            else:
                raise SaltInvocationError(
                    f"Could not resolve Subnet Name {subnet} to an ID."
                )
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    return _modify_resource(
        name,
        name_param="CacheSubnetGroupName",
        desc="cache subnet group",
        res_type="cache_subnet_group",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def delete_cache_subnet_group(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Delete an ElastiCache subnet group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.delete_subnet_group my-subnet-group region=us-east-1
    """
    return _delete_resource(
        name,
        name_param="CacheSubnetGroupName",
        desc="cache subnet group",
        res_type="cache_subnet_group",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def describe_cache_security_groups(
    name=None, conn=None, region=None, key=None, keyid=None, profile=None
):
    """
    Return details about all (or just one) Elasticache cache clusters.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.describe_cache_security_groups
        salt myminion boto3_elasticache.describe_cache_security_groups mycachesecgrp
    """
    return _describe_resource(
        name=name,
        name_param="CacheSecurityGroupName",
        res_type="cache_security_group",
        info_node="CacheSecurityGroups",
        conn=conn,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )


def cache_security_group_exists(name, region=None, key=None, keyid=None, profile=None):
    """
    Check to see if an ElastiCache security group exists.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.cache_security_group_exists mysecuritygroup
    """
    return bool(
        describe_cache_security_groups(
            name=name, region=region, key=key, keyid=keyid, profile=profile
        )
    )


def create_cache_security_group(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Create a cache security group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.create_cache_security_group mycachesecgrp Description='My Cache Security Group'
    """
    return _create_resource(
        name,
        name_param="CacheSecurityGroupName",
        desc="cache security group",
        res_type="cache_security_group",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def delete_cache_security_group(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Delete a cache security group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.delete_cache_security_group myelasticachesg
    """
    return _delete_resource(
        name,
        name_param="CacheSecurityGroupName",
        desc="cache security group",
        res_type="cache_security_group",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def authorize_cache_security_group_ingress(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Authorize network ingress from an ec2 security group to a cache security group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.authorize_cache_security_group_ingress \
                                        mycachesecgrp \
                                        EC2SecurityGroupName=someEC2sg \
                                        EC2SecurityGroupOwnerId=SOMEOWNERID
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if "CacheSecurityGroupName" in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided "
            "'CacheSecurityGroupName: %s'",
            name,
            args["CacheSecurityGroupName"],
        )
        name = args["CacheSecurityGroupName"]
    else:
        args["CacheSubnetGroupName"] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        conn.authorize_cache_security_group_ingress(**args)
        log.info(
            "Authorized %s to cache security group %s.",
            args["EC2SecurityGroupName"],
            name,
        )
        return True
    except botocore.exceptions.ClientError as e:
        log.error("Failed to update security group %s: %s", name, e)
        return False


def revoke_cache_security_group_ingress(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Revoke network ingress from an ec2 security group to a cache security
    group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.revoke_cache_security_group_ingress \
                                        mycachesecgrp \
                                        EC2SecurityGroupName=someEC2sg \
                                        EC2SecurityGroupOwnerId=SOMEOWNERID
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if "CacheSecurityGroupName" in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided "
            "'CacheSecurityGroupName: %s'",
            name,
            args["CacheSecurityGroupName"],
        )
        name = args["CacheSecurityGroupName"]
    else:
        args["CacheSubnetGroupName"] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        conn.revoke_cache_security_group_ingress(**args)
        log.info(
            "Revoked %s from cache security group %s.",
            args["EC2SecurityGroupName"],
            name,
        )
        return True
    except botocore.exceptions.ClientError as e:
        log.error("Failed to update security group %s: %s", name, e)
        return False


def list_tags_for_resource(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    List tags on an Elasticache resource.

    Note that this function is essentially useless as it requires a full AWS ARN for the
    resource being operated on, but there is no provided API or programmatic way to find
    the ARN for a given object from its name or ID alone.  It requires specific knowledge
    about the account number, AWS partition, and other magic details to generate.

    If you happen to have those handy, feel free to utilize this however...

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.list_tags_for_resource \
                name'=arn:aws:elasticache:us-west-2:0123456789:snapshot:mySnapshot'
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if "ResourceName" in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided "
            "'ResourceName: %s'",
            name,
            args["ResourceName"],
        )
        name = args["ResourceName"]
    else:
        args["ResourceName"] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        r = conn.list_tags_for_resource(**args)
        if r and "Taglist" in r:
            return r["TagList"]
        return []
    except botocore.exceptions.ClientError as e:
        log.error("Failed to list tags for resource %s: %s", name, e)
        return []


def add_tags_to_resource(name, region=None, key=None, keyid=None, profile=None, **args):
    """
    Add tags to an Elasticache resource.

    Note that this function is essentially useless as it requires a full AWS ARN for the
    resource being operated on, but there is no provided API or programmatic way to find
    the ARN for a given object from its name or ID alone.  It requires specific knowledge
    about the account number, AWS partition, and other magic details to generate.

    If you happen to have those at hand though, feel free to utilize this function...

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.add_tags_to_resource \
                name'=arn:aws:elasticache:us-west-2:0123456789:snapshot:mySnapshot' \
                Tags="[{'Key': 'TeamOwner', 'Value': 'infrastructure'}]"
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if "ResourceName" in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided "
            "'ResourceName: %s'",
            name,
            args["ResourceName"],
        )
        name = args["ResourceName"]
    else:
        args["ResourceName"] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        conn.add_tags_to_resource(**args)
        log.info("Added tags %s to %s.", args["Tags"], name)
        return True
    except botocore.exceptions.ClientError as e:
        log.error("Failed to add tags to %s: %s", name, e)
        return False


def remove_tags_from_resource(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Remove tags from an Elasticache resource.

    Note that this function is essentially useless as it requires a full AWS ARN for the
    resource being operated on, but there is no provided API or programmatic way to find
    the ARN for a given object from its name or ID alone.  It requires specific knowledge
    about the account number, AWS partition, and other magic details to generate.

    If you happen to have those at hand though, feel free to utilize this function...

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.remove_tags_from_resource \
                name'=arn:aws:elasticache:us-west-2:0123456789:snapshot:mySnapshot' \
                TagKeys="['TeamOwner']"
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if "ResourceName" in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided "
            "'ResourceName: %s'",
            name,
            args["ResourceName"],
        )
        name = args["ResourceName"]
    else:
        args["ResourceName"] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        conn.remove_tags_from_resource(**args)
        log.info("Added tags %s to %s.", args["Tags"], name)
        return True
    except botocore.exceptions.ClientError as e:
        log.error("Failed to add tags to %s: %s", name, e)
        return False


def copy_snapshot(name, region=None, key=None, keyid=None, profile=None, **args):
    """
    Make a copy of an existing snapshot.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.copy_snapshot name=mySnapshot \
                                                      TargetSnapshotName=copyOfMySnapshot
    """
    conn = _get_conn(region=region, key=key, keyid=keyid, profile=profile)
    if "SourceSnapshotName" in args:
        log.info(
            "'name: %s' param being overridden by explicitly provided "
            "'SourceSnapshotName: %s'",
            name,
            args["SourceSnapshotName"],
        )
        name = args["SourceSnapshotName"]
    else:
        args["SourceSnapshotName"] = name
    args = {k: v for k, v in args.items() if not k.startswith("_")}
    try:
        conn.copy_snapshot(**args)
        log.info("Snapshot %s copied to %s.", name, args["TargetSnapshotName"])
        return True
    except botocore.exceptions.ClientError as e:
        log.error("Failed to copy snapshot %s: %s", name, e)
        return False


def describe_cache_parameter_groups(
    name=None, conn=None, region=None, key=None, keyid=None, profile=None
):
    """
    Return details about all (or just one) Elasticache cache clusters.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.describe_cache_parameter_groups
        salt myminion boto3_elasticache.describe_cache_parameter_groups myParameterGroup
    """
    return _describe_resource(
        name=name,
        name_param="CacheParameterGroupName",
        res_type="cache_parameter_group",
        info_node="CacheParameterGroups",
        conn=conn,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )


def create_cache_parameter_group(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Create a cache parameter group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.create_cache_parameter_group \
                name=myParamGroup \
                CacheParameterGroupFamily=redis2.8 \
                Description="My Parameter Group"
    """
    return _create_resource(
        name,
        name_param="CacheParameterGroupName",
        desc="cache parameter group",
        res_type="cache_parameter_group",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )


def delete_cache_parameter_group(
    name, region=None, key=None, keyid=None, profile=None, **args
):
    """
    Delete a cache parameter group.

    Example:

    .. code-block:: bash

        salt myminion boto3_elasticache.delete_cache_parameter_group myParamGroup
    """
    return _delete_resource(
        name,
        name_param="CacheParameterGroupName",
        desc="cache parameter group",
        res_type="cache_parameter_group",
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
        **args,
    )



"""
    :codeauthor: `Gareth J. Greenaway <gareth@saltstack.com>`

    tests.unit.cloud.clouds.digitalocean_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import logging

import pytest

from salt.cloud.clouds import digitalocean
from salt.exceptions import SaltCloudSystemExit
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def vpc_list():
    return {
        "vpcs": [
            {
                "name": "env.prod-vpc",
                "description": "VPC for production environment",
                "region": "nyc1",
                "ip_range": "10.10.10.0/24",
                "id": "5a4981aa-9653-4bd1-bef5-d6bff52042e4",
                "urn": "do:vpc:5a4981aa-9653-4bd1-bef5-d6bff52042e4",
                "default": False,
                "created_at": "2020-03-13T19:20:47.442049222Z",
            },
            {
                "id": "e0fe0f4d-596a-465e-a902-571ce57b79fa",
                "urn": "do:vpc:e0fe0f4d-596a-465e-a902-571ce57b79fa",
                "name": "default-nyc1",
                "description": "",
                "region": "nyc1",
                "ip_range": "10.102.0.0/20",
                "created_at": "2020-03-13T19:29:20Z",
                "default": True,
            },
            {
                "id": "d455e75d-4858-4eec-8c95-da2f0a5f93a7",
                "urn": "do:vpc:d455e75d-4858-4eec-8c95-da2f0a5f93a7",
                "name": "default-nyc3",
                "description": "",
                "region": "nyc3",
                "ip_range": "10.100.0.0/20",
                "created_at": "2019-11-19T22:19:35Z",
                "default": True,
            },
        ],
        "links": {},
        "meta": {"total": 3},
    }


def test_reboot_no_call():
    """
    Tests that a SaltCloudSystemExit is raised when
    kwargs that are provided do not include an action.
    """
    with pytest.raises(SaltCloudSystemExit) as excinfo:
        digitalocean.reboot(name="fake_name")

    assert "The reboot action must be called with -a or --action." == str(excinfo.value)


def test__get_vpc_by_name(vpc_list):
    default_nyc1 = {
        "default-nyc1": {
            "default": True,
            "description": "",
            "id": "e0fe0f4d-596a-465e-a902-571ce57b79fa",
            "ip_range": "10.102.0.0/20",
            "name": "default-nyc1",
            "region": "nyc1",
            "urn": "do:vpc:e0fe0f4d-596a-465e-a902-571ce57b79fa",
        }
    }
    with patch(
        "salt.cloud.clouds.digitalocean.query", MagicMock(return_value=vpc_list)
    ):
        ret = digitalocean._get_vpc_by_name("default-nyc1")
        assert ret == default_nyc1
        ret = digitalocean._get_vpc_by_name("NOT-THERE")
        assert ret is None



import pytest

import salt.utils.path

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="module")
def cmd(modules):
    return modules.cmd


@pytest.fixture(scope="module")
def exitcode_script(state_tree):
    exit_code = 12345
    script_contents = f"""
    Write-Host "Expected exit code: {exit_code}"
    exit {exit_code}
    """
    with pytest.helpers.temp_file("exit_code.ps1", script_contents, state_tree):
        yield exit_code


@pytest.fixture(params=["powershell", "pwsh"])
def shell(request):
    """
    This will run the test on powershell and powershell core (pwsh). If
    powershell core is not installed that test run will be skipped
    """
    if request.param == "pwsh" and salt.utils.path.which("pwsh") is None:
        pytest.skip("Powershell 7 Not Present")
    return request.param


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account() as _account:
        yield _account


@pytest.fixture
def issue_56195(state_tree):
    contents = """
    [CmdLetBinding()]
    Param(
      [SecureString] $SecureString
    )
    $Credential = New-Object System.Net.NetworkCredential("DummyId", $SecureString)
    $Credential.Password
    """
    with pytest.helpers.temp_file("test.ps1", contents, state_tree / "issue-56195"):
        yield


def test_windows_script_args_powershell(cmd, shell, issue_56195):
    """
    Ensure that powershell processes an inline script with args where the args
    contain powershell that needs to be rendered
    """
    password = "i like cheese"
    args = (
        "-SecureString (ConvertTo-SecureString -String '{}' -AsPlainText -Force)"
        " -ErrorAction Stop".format(password)
    )
    script = "salt://issue-56195/test.ps1"

    ret = cmd.script(source=script, args=args, shell=shell, saltenv="base")

    assert ret["stdout"] == password


def test_windows_script_args_powershell_runas(cmd, shell, account, issue_56195):
    """
    Ensure that powershell processes an inline script with args where the args
    contain powershell that needs to be rendered
    """
    password = "i like cheese"
    args = (
        "-SecureString (ConvertTo-SecureString -String '{}' -AsPlainText -Force)"
        " -ErrorAction Stop".format(password)
    )
    script = "salt://issue-56195/test.ps1"

    ret = cmd.script(
        source=script,
        args=args,
        shell=shell,
        saltenv="base",
        runas=account.username,
        password=account.password,
    )

    assert ret["stdout"] == password


def test_windows_script_exitcode(cmd, shell, exitcode_script):
    ret = cmd.script("salt://exit_code.ps1", shell=shell, saltenv="base")
    assert ret["retcode"] == exitcode_script



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.rdp
"""

import pytest

import salt.modules.rdp as rdp
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rdp: {}}


# 'enable' function tests: 1


def test_enable():
    """
    Test if it enables RDP the service on the server
    """
    mock = MagicMock(return_value=True)
    with patch.dict(rdp.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.rdp._parse_return_code_powershell", MagicMock(return_value=0)
    ):
        assert rdp.enable()


# 'disable' function tests: 1


def test_disable():
    """
    Test if it disables RDP the service on the server
    """
    mock = MagicMock(return_value=True)
    with patch.dict(rdp.__salt__, {"cmd.run": mock}), patch(
        "salt.modules.rdp._parse_return_code_powershell", MagicMock(return_value=0)
    ):
        assert rdp.disable()


# 'status' function tests: 1


def test_status():
    """
    Test if it shows rdp is enabled on the server
    """
    mock = MagicMock(return_value="1")
    with patch.dict(rdp.__salt__, {"cmd.run": mock}):
        assert rdp.status()



"""
tests.integration.shell.cp
~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import logging
import os
import pathlib

import pytest

import salt.utils.files
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


@pytest.fixture
def source_testfile():
    src = pathlib.Path(RUNTIME_VARS.BASE_FILES) / "testfile"
    return str(src.resolve())


@pytest.fixture
def dest_testfile():
    dst = pathlib.Path(RUNTIME_VARS.TMP) / "test_cp_testfile_copy"
    try:
        yield str(dst)
    finally:
        if dst.exists():
            dst.unlink()


@pytest.mark.windows_whitelisted
@pytest.mark.core_test
def test_cp_testfile(salt_minion, salt_cp_cli, source_testfile, dest_testfile):
    """
    test salt-cp
    """
    ret = salt_cp_cli.run(source_testfile, dest_testfile, minion_tgt=salt_minion.id)
    assert ret.returncode == 0
    assert ret.data[dest_testfile] is True
    assert os.path.exists(dest_testfile)
    with salt.utils.files.fopen(source_testfile) as rfh:
        source_testfile_contents = rfh.read()
    with salt.utils.files.fopen(dest_testfile) as rfh:
        dest_test_file = rfh.read()
    assert source_testfile_contents == dest_test_file



"""
Initialize the engines system. This plugin system allows for
complex services to be encapsulated within the salt plugin environment
"""

import logging

import salt
import salt.loader
import salt.utils.platform
import salt.utils.process

log = logging.getLogger(__name__)


def start_engines(opts, proc_mgr, proxy=None):
    """
    Fire up the configured engines!
    """
    utils = salt.loader.utils(opts, proxy=proxy)
    if opts["__role"] == "master":
        runners = salt.loader.runner(opts, utils=utils)
    else:
        runners = []
    funcs = salt.loader.minion_mods(opts, utils=utils, proxy=proxy)
    engines = salt.loader.engines(opts, funcs, runners, utils, proxy=proxy)

    engines_opt = opts.get("engines", [])
    if isinstance(engines_opt, dict):
        engines_opt = [{k: v} for k, v in engines_opt.items()]

    # Function references are not picklable. Spawning platforms, Windows
    # and macOS, need to pickle when spawning, these will need to be recalculated
    # in the spawned child process.
    if salt.utils.platform.spawning_platform():
        runners = None
        utils = None
        funcs = None

    for engine in engines_opt:
        if isinstance(engine, dict):
            engine, engine_opts = next(iter(engine.items()))
        else:
            engine_opts = None
        engine_name = None
        if engine_opts is not None and "engine_module" in engine_opts:
            fun = "{}.start".format(engine_opts["engine_module"])
            engine_name = engine
            del engine_opts["engine_module"]
        else:
            fun = f"{engine}.start"
        if fun in engines:
            start_func = engines[fun]
            if engine_name:
                name = f"Engine({start_func.__module__}, name={engine_name})"
            else:
                name = f"Engine({start_func.__module__})"
            log.info("Starting %s", name)
            proc_mgr.add_process(
                Engine,
                args=(opts, fun, engine_opts, funcs, runners, proxy),
                name=name,
            )


class Engine(salt.utils.process.SignalHandlingProcess):
    """
    Execute the given engine in a new process
    """

    def __init__(self, opts, fun, config, funcs, runners, proxy, **kwargs):
        """
        Set up the process executor
        """
        super().__init__(**kwargs)
        self.opts = opts
        self.config = config
        self.fun = fun
        self.funcs = funcs
        self.runners = runners
        self.proxy = proxy

    def run(self):
        """
        Run the master service!
        """
        self.utils = salt.loader.utils(self.opts, proxy=self.proxy)
        if salt.utils.platform.spawning_platform():
            # Calculate function references since they can't be pickled.
            if self.opts["__role"] == "master":
                self.runners = salt.loader.runner(self.opts, utils=self.utils)
            else:
                self.runners = []
            self.funcs = salt.loader.minion_mods(
                self.opts, utils=self.utils, proxy=self.proxy
            )

        self.engine = salt.loader.engines(
            self.opts, self.funcs, self.runners, self.utils, proxy=self.proxy
        )
        kwargs = self.config or {}

        try:
            self.engine[self.fun](**kwargs)
        except Exception:  # pylint: disable=broad-except
            log.critical(
                "%s could not be started!",
                self.name,
                exc_info=True,
            )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.aws_sqs as aws_sqs
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {aws_sqs: {}}


def test_exists():
    """
    Test to ensure the SQS queue exists.
    """
    name = "myqueue"
    region = "eu-west-1"

    ret = {"name": name, "result": None, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(aws_sqs.__salt__, {"aws_sqs.queue_exists": mock}):
        comt = f"AWS SQS queue {name} is set to be created"
        ret.update({"comment": comt})
        with patch.dict(aws_sqs.__opts__, {"test": True}):
            assert aws_sqs.exists(name, region) == ret

        comt = f"{name} exists in {region}"
        ret.update({"comment": comt, "result": True})
        assert aws_sqs.exists(name, region) == ret


def test_absent():
    """
    Test to remove the named SQS queue if it exists.
    """
    name = "myqueue"
    region = "eu-west-1"

    ret = {"name": name, "result": None, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, False])
    with patch.dict(aws_sqs.__salt__, {"aws_sqs.queue_exists": mock}):
        comt = f"AWS SQS queue {name} is set to be removed"
        ret.update({"comment": comt})
        with patch.dict(aws_sqs.__opts__, {"test": True}):
            assert aws_sqs.absent(name, region) == ret

        comt = f"{name} does not exist in {region}"
        ret.update({"comment": comt, "result": True})
        assert aws_sqs.absent(name, region) == ret



import asyncio
import multiprocessing
import socket
import threading
import time

import pytest

import salt.transport.tcp

pytestmark = [
    pytest.mark.core_test,
]


@pytest.mark.skip_on_fips_enabled_platform
def test_tcp_load_balancer_server(master_opts, io_loop):

    messages = []

    def handler(stream, message, header):
        messages.append(message)

    queue = multiprocessing.Queue()
    server = salt.transport.tcp.LoadBalancerServer(master_opts, queue)
    worker = salt.transport.tcp.LoadBalancerWorker(queue, handler, io_loop=io_loop)

    def run_loop():
        try:
            io_loop.start()
        except Exception as exc:  # pylint: disable=broad-except
            print(f"Caught exeption {exc}")

    thread = threading.Thread(target=server.run)
    thread.start()

    # Wait for bind to happen.
    time.sleep(0.5)

    package = {"foo": "bar"}
    payload = salt.transport.frame.frame_msg(package)
    sock = socket.socket()
    sock.connect(("127.0.0.1", master_opts["ret_port"]))
    sock.send(payload)

    start = time.monotonic()

    async def check_test():
        while not messages:
            await asyncio.sleep(0.3)
            if time.monotonic() - start > 30:
                break

    io_loop.run_sync(lambda: check_test())  # pylint: disable=unnecessary-lambda

    try:
        if time.monotonic() - start > 30:
            assert False, "Took longer than 30 seconds to receive message"

        assert [package] == messages
    finally:
        server.close()
        thread.join()
        worker.close()



"""
:maintainer:    Alberto Planas <aplanas@suse.com>
:platform:      Linux
"""

import sys

import pytest

import salt.loader.context
import salt.modules.chroot as chroot
import salt.utils.platform
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skip_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    loader_context = salt.loader.context.LoaderContext()
    return {
        chroot: {
            "__salt__": {},
            "__utils__": {"files.rm_rf": MagicMock()},
            "__opts__": {"extension_modules": "", "cachedir": "/tmp/"},
            "__pillar__": salt.loader.context.NamedLoaderContext(
                "__pillar__", loader_context, {}
            ),
        }
    }


def test__create_and_execute_salt_state():
    with patch("salt.client.ssh.wrapper.state._cleanup_slsmod_low_data", MagicMock()):
        with patch("salt.utils.hashutils.get_hash", MagicMock(return_value="deadbeaf")):
            with patch("salt.fileclient.get_file_client", MagicMock()):
                with patch("salt.modules.chroot.call", MagicMock()):
                    chroot._create_and_execute_salt_state("", {}, {}, False, "md5")


def test_exist():
    """
    Test if the chroot environment exist.
    """
    with patch("os.path.isdir") as isdir:
        isdir.side_effect = (True, True, True, True)
        assert chroot.exist("/chroot")

    with patch("os.path.isdir") as isdir:
        isdir.side_effect = (True, True, True, False)
        assert not chroot.exist("/chroot")


def test_create():
    """
    Test the creation of an empty chroot environment.
    """
    with patch("os.makedirs") as makedirs:
        with patch("salt.modules.chroot.exist") as exist:
            exist.return_value = True
            assert chroot.create("/chroot")
            makedirs.assert_not_called()

    with patch("os.makedirs") as makedirs:
        with patch("salt.modules.chroot.exist") as exist:
            exist.return_value = False
            assert chroot.create("/chroot")
            makedirs.assert_called()


def test_in_chroot():
    """
    Test the detection of chroot environment.
    """
    matrix = (("a", "b", True), ("a", "a", False))
    with patch("salt.utils.files.fopen") as fopen:
        for root_mountinfo, self_mountinfo, result in matrix:
            fopen.return_value.__enter__.return_value = fopen
            fopen.read = MagicMock(side_effect=(root_mountinfo, self_mountinfo))
            assert chroot.in_chroot() == result


def test_call_fails_input_validation():
    """
    Test execution of Salt functions in chroot.
    """
    # Basic input validation
    with patch("salt.modules.chroot.exist") as exist:
        exist.return_value = False
        pytest.raises(CommandExecutionError, chroot.call, "/chroot", "")
        pytest.raises(CommandExecutionError, chroot.call, "/chroot", "test.ping")


def test_call_fails_untar():
    """
    Test execution of Salt functions in chroot.
    """
    # Fail the tar command
    with patch("salt.modules.chroot.exist") as exist:
        with patch("tempfile.mkdtemp") as mkdtemp:
            exist.return_value = True
            mkdtemp.return_value = "/chroot/tmp01"
            utils_mock = {
                "thin.gen_thin": MagicMock(return_value="/salt-thin.tgz"),
                "files.rm_rf": MagicMock(),
            }
            salt_mock = {
                "cmd.run": MagicMock(return_value="Error"),
                "config.option": MagicMock(),
            }
            with patch.dict(chroot.__utils__, utils_mock), patch.dict(
                chroot.__salt__, salt_mock
            ):
                assert chroot.call("/chroot", "test.ping") == {
                    "result": False,
                    "comment": "Error",
                }
                utils_mock["thin.gen_thin"].assert_called_once()
                salt_mock["config.option"].assert_called()
                salt_mock["cmd.run"].assert_called_once()
                utils_mock["files.rm_rf"].assert_called_once()


def test_call_fails_salt_thin():
    """
    Test execution of Salt functions in chroot.
    """
    # Fail the inner command
    with patch("salt.modules.chroot.exist") as exist:
        with patch("tempfile.mkdtemp") as mkdtemp:
            exist.return_value = True
            mkdtemp.return_value = "/chroot/tmp01"
            utils_mock = {
                "thin.gen_thin": MagicMock(return_value="/salt-thin.tgz"),
                "files.rm_rf": MagicMock(),
                "json.find_json": MagicMock(side_effect=ValueError()),
            }
            salt_mock = {
                "cmd.run": MagicMock(return_value=""),
                "config.option": MagicMock(),
                "cmd.run_chroot": MagicMock(
                    return_value={"retcode": 1, "stdout": "", "stderr": "Error"}
                ),
            }
            with patch.dict(chroot.__utils__, utils_mock), patch.dict(
                chroot.__salt__, salt_mock
            ):
                assert chroot.call("/chroot", "test.ping") == {
                    "result": False,
                    "retcode": 1,
                    "comment": {"stdout": "", "stderr": "Error"},
                }
                utils_mock["thin.gen_thin"].assert_called_once()
                salt_mock["config.option"].assert_called()
                salt_mock["cmd.run"].assert_called_once()
                salt_mock["cmd.run_chroot"].assert_called_with(
                    "/chroot",
                    [
                        f"python{sys.version_info[0]}",
                        "/tmp01/salt-call",
                        "--metadata",
                        "--local",
                        "--log-file",
                        "/tmp01/log",
                        "--cachedir",
                        "/tmp01/cache",
                        "--out",
                        "json",
                        "-l",
                        "quiet",
                        "--",
                        "test.ping",
                    ],
                )
                utils_mock["files.rm_rf"].assert_called_once()


def test_call_success():
    """
    Test execution of Salt functions in chroot.
    """
    # Success test
    with patch("salt.modules.chroot.exist") as exist:
        with patch("tempfile.mkdtemp") as mkdtemp:
            exist.return_value = True
            mkdtemp.return_value = "/chroot/tmp01"
            utils_mock = {
                "thin.gen_thin": MagicMock(return_value="/salt-thin.tgz"),
                "files.rm_rf": MagicMock(),
                "json.find_json": MagicMock(return_value={"return": "result"}),
            }
            salt_mock = {
                "cmd.run": MagicMock(return_value=""),
                "config.option": MagicMock(),
                "cmd.run_chroot": MagicMock(return_value={"retcode": 0, "stdout": ""}),
            }
            with patch.dict(chroot.__utils__, utils_mock), patch.dict(
                chroot.__salt__, salt_mock
            ):
                assert chroot.call("/chroot", "test.ping") == "result"
                utils_mock["thin.gen_thin"].assert_called_once()
                salt_mock["config.option"].assert_called()
                salt_mock["cmd.run"].assert_called_once()
                salt_mock["cmd.run_chroot"].assert_called_with(
                    "/chroot",
                    [
                        f"python{sys.version_info[0]}",
                        "/tmp01/salt-call",
                        "--metadata",
                        "--local",
                        "--log-file",
                        "/tmp01/log",
                        "--cachedir",
                        "/tmp01/cache",
                        "--out",
                        "json",
                        "-l",
                        "quiet",
                        "--",
                        "test.ping",
                    ],
                )
                utils_mock["files.rm_rf"].assert_called_once()


def test_call_success_parameters():
    """
    Test execution of Salt functions in chroot with parameters.
    """
    # Success test
    with patch("salt.modules.chroot.exist") as exist:
        with patch("tempfile.mkdtemp") as mkdtemp:
            exist.return_value = True
            mkdtemp.return_value = "/chroot/tmp01"
            utils_mock = {
                "thin.gen_thin": MagicMock(return_value="/salt-thin.tgz"),
                "files.rm_rf": MagicMock(),
                "json.find_json": MagicMock(return_value={"return": "result"}),
            }
            salt_mock = {
                "cmd.run": MagicMock(return_value=""),
                "config.option": MagicMock(),
                "cmd.run_chroot": MagicMock(return_value={"retcode": 0, "stdout": ""}),
            }
            with patch.dict(chroot.__utils__, utils_mock), patch.dict(
                chroot.__salt__, salt_mock
            ):
                assert (
                    chroot.call("/chroot", "module.function", key="value") == "result"
                )
                utils_mock["thin.gen_thin"].assert_called_once()
                salt_mock["config.option"].assert_called()
                salt_mock["cmd.run"].assert_called_once()
                salt_mock["cmd.run_chroot"].assert_called_with(
                    "/chroot",
                    [
                        f"python{sys.version_info[0]}",
                        "/tmp01/salt-call",
                        "--metadata",
                        "--local",
                        "--log-file",
                        "/tmp01/log",
                        "--cachedir",
                        "/tmp01/cache",
                        "--out",
                        "json",
                        "-l",
                        "quiet",
                        "--",
                        "module.function",
                        "key=value",
                    ],
                )
                utils_mock["files.rm_rf"].assert_called_once()


def test_sls():
    """
    Test execution of Salt states in chroot.
    """
    with patch("salt.utils.state.get_sls_opts") as get_sls_opts:
        with patch("salt.fileclient.get_file_client") as get_file_client:
            with patch("salt.client.ssh.state.SSHHighState") as SSHHighState:
                with patch(
                    "salt.modules.chroot._create_and_execute_salt_state"
                ) as _create_and_execute_salt_state:
                    SSHHighState.return_value = SSHHighState
                    SSHHighState.render_highstate.return_value = (None, [])
                    SSHHighState.state.reconcile_extend.return_value = (None, [])
                    SSHHighState.state.requisite_in.return_value = (None, [])
                    SSHHighState.state.verify_high.return_value = []

                    _create_and_execute_salt_state.return_value = "result"
                    opts_mock = {
                        "hash_type": "md5",
                    }
                    get_sls_opts.return_value = opts_mock
                    with patch.dict(chroot.__opts__, opts_mock):
                        assert chroot.sls("/chroot", "module") == "result"
                        _create_and_execute_salt_state.assert_called_once()


def test_highstate():
    """
    Test execution of Salt states in chroot.
    """
    with patch("salt.utils.state.get_sls_opts") as get_sls_opts:
        with patch("salt.fileclient.get_file_client") as get_file_client:
            with patch("salt.client.ssh.state.SSHHighState") as SSHHighState:
                with patch(
                    "salt.modules.chroot._create_and_execute_salt_state"
                ) as _create_and_execute_salt_state:
                    SSHHighState.return_value = SSHHighState

                    _create_and_execute_salt_state.return_value = "result"
                    opts_mock = {
                        "hash_type": "md5",
                    }
                    get_sls_opts.return_value = opts_mock
                    with patch.dict(chroot.__opts__, opts_mock):
                        assert chroot.highstate("/chroot") == "result"
                        _create_and_execute_salt_state.assert_called_once()



import pytest

import salt.modules.influxdbmod as influx_mod
import salt.states.influxdb_continuous_query as influx
from tests.support.mock import create_autospec, patch


@pytest.fixture
def configure_loader_modules():
    return {influx: {"__salt__": {}, "__opts__": {"test": False}}}


@pytest.mark.parametrize(
    "expected_kwargs",
    (
        {},
        {"something": "extra"},
        {"something": "extra", "even": "more"},
        {"something": "extra", "still": "more and more and more", "and": "more"},
        {
            "something": "extra",
            "what": "in tarnation",
            "do": "you want",
            "to": "add here?",
        },
    ),
)
def test_when_present_is_called_it_should_pass_client_args_to_create_module(
    expected_kwargs,
):
    influx_module = create_autospec(influx_mod)
    influx_module.continuous_query_exists.return_value = False
    with patch.dict(
        influx.__salt__,
        {
            "influxdb.continuous_query_exists": influx_module.continuous_query_exists,
            "influxdb.create_continuous_query": influx_module.create_continuous_query,
        },
    ):
        influx.present(
            name="foo",
            database="fnord",
            query="fnord",
            resample_time="whatever",
            coverage_period="fnord",
            **expected_kwargs
        )

    actual_kwargs = influx_module.create_continuous_query.mock_calls[0].kwargs

    assert actual_kwargs == expected_kwargs



import pytest

import salt.config
import salt.wheel


@pytest.fixture
def client_config(salt_master):
    config = salt.config.client_config(
        salt_master.config["conf_file"],
        defaults=salt_master.config.copy(),
    )
    return config


@pytest.fixture
def client(client_config):
    return salt.wheel.Wheel(client_config)


@pytest.fixture
def auth_creds(salt_auto_account):
    return {
        "username": salt_auto_account.username,
        "password": salt_auto_account.password,
        "eauth": "auto",
    }



"""
    :codeauthor: :email:`Simon Dodsley <simon@purestorage.com>`
"""

import pytest

import salt.modules.purefb as purefb
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {purefb: {}}


def test_fs_create():
    """
    Test for creation of a filesystem
    """
    with patch.object(purefb, "fs_create", return_value=True):
        assert purefb.fs_create("test") is True


def test_fs_delete():
    """
    Test for deletion of a filesystem
    """
    with patch.object(purefb, "fs_delete", return_value=True):
        assert purefb.fs_delete("test") is True


def test_fs_eradicate():
    """
    Test for eradication of a filesystem
    """
    with patch.object(purefb, "fs_eradicate", return_value=True):
        assert purefb.fs_eradicate("test") is True


def test_fs_extend():
    """
    Test for size extention of a filesystem
    """
    with patch.object(purefb, "fs_extend", return_value=True):
        assert purefb.fs_extend("test", "33G") is True


def test_snap_create():
    """
    Test for creation of a filesystem snapshot
    """
    with patch.object(purefb, "snap_create", return_value=True):
        assert purefb.snap_create("test", suffix="suffix") is True


def test_snap_delete():
    """
    Test for deletion of a filesystem snapshot
    """
    with patch.object(purefb, "snap_delete", return_value=True):
        assert purefb.snap_delete("test", suffix="suffix") is True


def test_snap_eradicate():
    """
    Test for eradication of a deleted filesystem snapshot
    """
    with patch.object(purefb, "snap_eradicate", return_value=True):
        assert purefb.snap_eradicate("test", suffix="suffix") is True



import pytest

pytestmark = [pytest.mark.skip_unless_on_linux]


@pytest.fixture(scope="module")
def defaults(modules):
    return modules.defaults


def test_defaults_get(defaults, state_tree, caplog):
    """
    test defaults.get
    """

    json_contents = """
    {
        "users": {
            "root": 1
        }
    }
    """
    path = state_tree / "core"
    with pytest.helpers.temp_file("defaults.json", json_contents, path):
        assert defaults.get("core:users:root") == 1


def test_defaults_merge(defaults):
    """
    test defaults.merge
    """
    assert defaults.merge({"a": "b"}, {"d": "e"}) == {"a": "b", "d": "e"}


def test_defaults_deepcopy(defaults):
    """
    test defaults.deepcopy
    """
    test_dict = {"1": "one"}
    assert defaults.deepcopy(test_dict) == test_dict



"""
Manage RDSs
===========

.. versionadded:: 2015.8.0

Create and destroy RDS instances. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses ``boto``, which can be installed via package, or pip.

This module accepts explicit rds credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    rds.keyid: GKTADJGHEIQSXMKKRBJ08H
    rds.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile,
either passed in as a dict, or as a string to pull from pillars or minion
config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

.. code-block:: yaml

    Ensure myrds RDS exists:
      boto_rds.present:
        - name: myrds
        - allocated_storage: 5
        - storage_type: standard
        - db_instance_class: db.t2.micro
        - engine: MySQL
        - master_username: myuser
        - master_user_password: mypass
        - region: us-east-1
        - keyid: GKTADJGHEIQSXMKKRBJ08H
        - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        - tags:
            key: value

.. code-block:: yaml

    Ensure parameter group exists:
        create-parameter-group:
          boto_rds.parameter_present:
            - name: myparametergroup
            - db_parameter_group_family: mysql5.6
            - description: "parameter group family"
            - parameters:
              - binlog_cache_size: 32768
              - binlog_checksum: CRC32
            - region: eu-west-1

:depends: boto3

"""

import logging
import os

import salt.utils.data
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto is available.
    """
    if "boto_rds.exists" in __salt__:
        return "boto_rds"
    return (False, __salt__.missing_fun_string("boto_rds.exists"))


def present(
    name,
    allocated_storage,
    db_instance_class,
    engine,
    master_username,
    master_user_password,
    db_name=None,
    storage_type=None,
    db_security_groups=None,
    vpc_security_group_ids=None,
    vpc_security_groups=None,
    availability_zone=None,
    db_subnet_group_name=None,
    preferred_maintenance_window=None,
    db_parameter_group_name=None,
    db_cluster_identifier=None,
    tde_credential_arn=None,
    tde_credential_password=None,
    storage_encrypted=None,
    kms_keyid=None,
    backup_retention_period=None,
    preferred_backup_window=None,
    port=None,
    multi_az=None,
    engine_version=None,
    auto_minor_version_upgrade=None,
    license_model=None,
    iops=None,
    option_group_name=None,
    character_set_name=None,
    publicly_accessible=None,
    wait_status=None,
    tags=None,
    copy_tags_to_snapshot=None,
    region=None,
    domain=None,
    key=None,
    keyid=None,
    monitoring_interval=None,
    monitoring_role_arn=None,
    domain_iam_role_name=None,
    promotion_tier=None,
    profile=None,
):
    """
    Ensure RDS instance exists.

    name
        Name of the RDS state definition.

    allocated_storage
        The amount of storage (in gigabytes) to be initially allocated for the
        database instance.

    db_instance_class
        The compute and memory capacity of the Amazon RDS DB instance.

    engine
        The name of the database engine to be used for this instance. Supported
        engine types are: MySQL, mariadb, oracle-se1, oracle-se, oracle-ee, sqlserver-ee,
        sqlserver-se, sqlserver-ex, sqlserver-web, postgres and aurora. For more
        information, please see the ``engine`` argument in the Boto3 RDS
        `create_db_instance`_ documentation.

    master_username
        The name of master user for the client DB instance.

    master_user_password
        The password for the master database user. Can be any printable ASCII
        character except "/", '"', or "@".

    db_name
        The meaning of this parameter differs according to the database engine you use.
        See the Boto3 RDS documentation to determine the appropriate value for your configuration.
        https://boto3.readthedocs.io/en/latest/reference/services/rds.html#RDS.Client.create_db_instance

    storage_type
        Specifies the storage type to be associated with the DB instance.
        Options are standard, gp2 and io1. If you specify io1, you must also include
        a value for the Iops parameter.

    db_security_groups
        A list of DB security groups to associate with this DB instance.

    vpc_security_group_ids
        A list of EC2 VPC security group IDs to associate with this DB instance.

    vpc_security_groups
        A list of EC2 VPC security groups (IDs or Name tags) to associate with this DB instance.

    availability_zone
        The EC2 Availability Zone that the database instance will be created
        in.

    db_subnet_group_name
        A DB subnet group to associate with this DB instance.

    preferred_maintenance_window
        The weekly time range (in UTC) during which system maintenance can
        occur.

    db_parameter_group_name
        A DB parameter group to associate with this DB instance.

    db_cluster_identifier
        If the DB instance is a member of a DB cluster, contains the name of
        the DB cluster that the DB instance is a member of.

    tde_credential_arn
        The ARN from the Key Store with which the instance is associated for
        TDE encryption.

    tde_credential_password
        The password to use for TDE encryption if an encryption key is not used.

    storage_encrypted
        Specifies whether the DB instance is encrypted.

    kms_keyid
        If storage_encrypted is true, the KMS key identifier for the encrypted
        DB instance.

    backup_retention_period
        The number of days for which automated backups are retained.

    preferred_backup_window
        The daily time range during which automated backups are created if
        automated backups are enabled.

    port
        The port number on which the database accepts connections.

    multi_az
        Specifies if the DB instance is a Multi-AZ deployment. You cannot set
        the AvailabilityZone parameter if the MultiAZ parameter is set to true.

    engine_version
        The version number of the database engine to use.

    auto_minor_version_upgrade
        Indicates that minor engine upgrades will be applied automatically to
        the DB instance during the maintenance window.

    license_model
        License model information for this DB instance.

    iops
        The amount of Provisioned IOPS (input/output operations per second) to
        be initially allocated for the DB instance.

    option_group_name
        Indicates that the DB instance should be associated with the specified
        option group.

    character_set_name
        For supported engines, indicates that the DB instance should be
        associated with the specified CharacterSet.

    publicly_accessible
        Specifies the accessibility options for the DB instance. A value of
        true specifies an Internet-facing instance with a publicly resolvable
        DNS name, which resolves to a public IP address. A value of false
        specifies an internal instance with a DNS name that resolves to a
        private IP address.

    wait_status
        Wait for the RDS instance to reach a desired status before finishing
        the state. Available states: available, modifying, backing-up

    tags
        A dict of tags.

    copy_tags_to_snapshot
        Specifies whether tags are copied from the DB instance to snapshots of
        the DB instance.

    region
        Region to connect to.

    domain
        The identifier of the Active Directory Domain.

    key
        AWS secret key to be used.

    keyid
        AWS access key to be used.

    monitoring_interval
        The interval, in seconds, between points when Enhanced Monitoring
        metrics are collected for the DB instance.

    monitoring_role_arn
        The ARN for the IAM role that permits RDS to send Enhanced Monitoring
        metrics to CloudWatch Logs.

    domain_iam_role_name
        Specify the name of the IAM role to be used when making API calls to
        the Directory Service.

    promotion_tier
        A value that specifies the order in which an Aurora Replica is
        promoted to the primary instance after a failure of the existing
        primary instance. For more information, see Fault Tolerance for an
        Aurora DB Cluster .

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.

    .. _create_db_instance: https://boto3.readthedocs.io/en/latest/reference/services/rds.html#RDS.Client.create_db_instance
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    r = __salt__["boto_rds.exists"](name, tags, region, key, keyid, profile)

    if not r.get("exists"):
        if __opts__["test"]:
            ret["comment"] = f"RDS instance {name} would be created."
            ret["result"] = None
            return ret

        r = __salt__["boto_rds.create"](
            name,
            allocated_storage,
            db_instance_class,
            engine,
            master_username,
            master_user_password,
            db_name,
            db_security_groups,
            vpc_security_group_ids,
            vpc_security_groups,
            availability_zone,
            db_subnet_group_name,
            preferred_maintenance_window,
            db_parameter_group_name,
            backup_retention_period,
            preferred_backup_window,
            port,
            multi_az,
            engine_version,
            auto_minor_version_upgrade,
            license_model,
            iops,
            option_group_name,
            character_set_name,
            publicly_accessible,
            wait_status,
            tags,
            db_cluster_identifier,
            storage_type,
            tde_credential_arn,
            tde_credential_password,
            storage_encrypted,
            kms_keyid,
            domain,
            copy_tags_to_snapshot,
            monitoring_interval,
            monitoring_role_arn,
            domain_iam_role_name,
            region,
            promotion_tier,
            key,
            keyid,
            profile,
        )

        if not r.get("created"):
            ret["result"] = False
            ret["comment"] = "Failed to create RDS instance {}.".format(
                r["error"]["message"]
            )
            return ret

        ret["changes"]["old"] = {"instance": None}
        ret["changes"]["new"] = {
            "instance": __salt__["boto_rds.describe_db_instances"](
                name=name,
                jmespath="DBInstances[0]",
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
        }
        ret["comment"] = f"RDS instance {name} created."
    else:
        ret["comment"] = f"RDS instance {name} exists."
    return ret


def replica_present(
    name,
    source,
    db_instance_class=None,
    availability_zone=None,
    port=None,
    auto_minor_version_upgrade=None,
    iops=None,
    option_group_name=None,
    publicly_accessible=None,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
    db_parameter_group_name=None,
):
    """
    Ensure RDS replica exists.

    .. code-block:: yaml

        Ensure myrds replica RDS exists:
          boto_rds.create_replica:
            - name: myreplica
            - source: mydb
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    replica_exists = __salt__["boto_rds.exists"](
        name, tags, region, key, keyid, profile
    )
    if not replica_exists.get("exists"):
        if __opts__["test"]:
            ret["comment"] = f"RDS read replica {name} is set to be created "
            ret["result"] = None
            return ret
        created = __salt__["boto_rds.create_read_replica"](
            name,
            source,
            db_instance_class,
            availability_zone,
            port,
            auto_minor_version_upgrade,
            iops,
            option_group_name,
            publicly_accessible,
            tags,
            region,
            key,
            keyid,
            profile,
        )
        if created:
            ret["comment"] = f"RDS replica {name} created."
            ret["changes"]["old"] = {"instance": None}
            ret["changes"]["new"] = {
                "instance": __salt__["boto_rds.describe_db_instances"](
                    name=name,
                    jmespath="DBInstances[0]",
                    region=region,
                    key=key,
                    keyid=keyid,
                    profile=profile,
                )
            }
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to create RDS replica {name}."
    else:
        jmespath = "DBInstances[0].DBParameterGroups[0].DBParameterGroupName"
        pmg_name = __salt__["boto_rds.describe_db_instances"](
            name=name,
            jmespath=jmespath,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        pmg_name = pmg_name[0] if pmg_name else None
        if pmg_name != db_parameter_group_name:
            modified = __salt__["boto_rds.modify_db_instance"](
                name=name,
                db_parameter_group_name=db_parameter_group_name,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if not modified:
                ret["result"] = False
                ret["comment"] = (
                    f"Failed to update parameter group of {name} RDS instance."
                )
            ret["changes"]["old"] = pmg_name
            ret["changes"]["new"] = db_parameter_group_name
        ret["result"] = True
        ret["comment"] = f"RDS replica {name} exists."
    return ret


def subnet_group_present(
    name,
    description,
    subnet_ids=None,
    subnet_names=None,
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure DB subnet group exists.

    name
        The name for the DB subnet group. This value is stored as a lowercase string.

    subnet_ids
        A list of the EC2 Subnet IDs for the DB subnet group.
        Either subnet_ids or subnet_names must be provided.

    subnet_names
        A list of The EC2 Subnet names for the DB subnet group.
        Either subnet_ids or subnet_names must be provided.

    description
        Subnet group description.

    tags
        A dict of tags.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    """
    if not salt.utils.data.exactly_one((subnet_ids, subnet_names)):
        raise SaltInvocationError(
            "One (but not both) of subnet_ids or subnet_names must be provided."
        )

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if not subnet_ids:
        subnet_ids = []

    if subnet_names:
        for i in subnet_names:
            r = __salt__["boto_vpc.get_resource_id"](
                "subnet", name=i, region=region, key=key, keyid=keyid, profile=profile
            )

            if "error" in r:
                ret["comment"] = "Error looking up subnet ids: {}".format(
                    r["error"]["message"]
                )
                ret["result"] = False
                return ret
            if r["id"] is None:
                ret["comment"] = f"Subnet {i} does not exist."
                ret["result"] = False
                return ret
            subnet_ids.append(r["id"])

    exists = __salt__["boto_rds.subnet_group_exists"](
        name=name, tags=tags, region=region, key=key, keyid=keyid, profile=profile
    )
    if not exists.get("exists"):
        if __opts__["test"]:
            ret["comment"] = f"Subnet group {name} is set to be created."
            ret["result"] = None
            return ret
        created = __salt__["boto_rds.create_subnet_group"](
            name=name,
            description=description,
            subnet_ids=subnet_ids,
            tags=tags,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )

        if not created:
            ret["result"] = False
            ret["comment"] = f"Failed to create {name} subnet group."
            return ret
        ret["changes"]["old"] = None
        ret["changes"]["new"] = name
        ret["comment"] = f"Subnet {name} created."
        return ret
    else:
        ret["comment"] = f"Subnet {name} present."

        return ret


def absent(
    name,
    skip_final_snapshot=None,
    final_db_snapshot_identifier=None,
    tags=None,
    wait_for_deletion=True,
    timeout=180,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure RDS instance is absent.

    name
        Name of the RDS instance.

    skip_final_snapshot
        Whether a final db snapshot is created before the instance is deleted.
        If True, no snapshot is created.
        If False, a snapshot is created before deleting the instance.

    final_db_snapshot_identifier
        If a final snapshot is requested, this is the identifier used for that
        snapshot.

    tags
        A dict of tags.

    wait_for_deletion (bool)
        Wait for the RDS instance to be deleted completely before finishing
        the state.

    timeout (in seconds)
        The amount of time that can pass before raising an Exception.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    current = __salt__["boto_rds.describe_db_instances"](
        name=name, region=region, key=key, keyid=keyid, profile=profile
    )
    if not current:
        ret["result"] = True
        ret["comment"] = f"{name} RDS already absent."
        return ret

    if __opts__["test"]:
        ret["comment"] = f"RDS {name} would be removed."
        ret["result"] = None
        return ret
    deleted = __salt__["boto_rds.delete"](
        name,
        skip_final_snapshot,
        final_db_snapshot_identifier,
        region,
        key,
        keyid,
        profile,
        tags,
        wait_for_deletion,
        timeout,
    )
    if not deleted:
        ret["result"] = False
        ret["comment"] = f"Failed to delete {name} RDS."
        return ret
    ret["changes"]["old"] = {"instance": current[0]}
    ret["changes"]["new"] = {"instance": None}
    ret["comment"] = f"RDS {name} deleted."
    return ret


def subnet_group_absent(
    name, tags=None, region=None, key=None, keyid=None, profile=None
):
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    exists = __salt__["boto_rds.subnet_group_exists"](
        name=name, tags=tags, region=region, key=key, keyid=keyid, profile=profile
    )
    if not exists:
        ret["result"] = True
        ret["comment"] = f"{name} RDS subnet group does not exist."
        return ret

    if __opts__["test"]:
        ret["comment"] = f"RDS subnet group {name} is set to be removed."
        ret["result"] = None
        return ret
    deleted = __salt__["boto_rds.delete_subnet_group"](
        name, region, key, keyid, profile
    )
    if not deleted:
        ret["result"] = False
        ret["comment"] = f"Failed to delete {name} RDS subnet group."
        return ret
    ret["changes"]["old"] = name
    ret["changes"]["new"] = None
    ret["comment"] = f"RDS subnet group {name} deleted."
    return ret


def parameter_present(
    name,
    db_parameter_group_family,
    description,
    parameters=None,
    apply_method="pending-reboot",
    tags=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure DB parameter group exists and update parameters.

    name
        The name for the parameter group.

    db_parameter_group_family
        The DB parameter group family name. A
        DB parameter group can be associated with one and only one DB
        parameter group family, and can be applied only to a DB instance
        running a database engine and engine version compatible with that
        DB parameter group family.

    description
        Parameter group description.

    parameters
        The DB parameters that need to be changed of type dictionary.

    apply_method
        The `apply-immediate` method can be used only for dynamic
        parameters; the `pending-reboot` method can be used with MySQL
        and Oracle DB instances for either dynamic or static
        parameters. For Microsoft SQL Server DB instances, the
        `pending-reboot` method can be used only for static
        parameters.

    tags
        A dict of tags.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}
    res = __salt__["boto_rds.parameter_group_exists"](
        name=name, tags=tags, region=region, key=key, keyid=keyid, profile=profile
    )
    if not res.get("exists"):
        if __opts__["test"]:
            ret["comment"] = f"Parameter group {name} is set to be created."
            ret["result"] = None
            return ret
        created = __salt__["boto_rds.create_parameter_group"](
            name=name,
            db_parameter_group_family=db_parameter_group_family,
            description=description,
            tags=tags,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        )
        if not created:
            ret["result"] = False
            ret["comment"] = f"Failed to create {name} parameter group."
            return ret
        ret["changes"]["New Parameter Group"] = name
        ret["comment"] = f"Parameter group {name} created."
    else:
        ret["comment"] = f"Parameter group {name} present."
    if parameters is not None:
        params = {}
        changed = {}
        for items in parameters:
            for k, value in items.items():
                if type(value) is bool:
                    params[k] = "on" if value else "off"
                else:
                    params[k] = str(value)
        log.debug("Parameters from user are : %s.", params)
        options = __salt__["boto_rds.describe_parameters"](
            name=name, region=region, key=key, keyid=keyid, profile=profile
        )
        if not options.get("result"):
            ret["result"] = False
            ret["comment"] = os.linesep.join(
                [ret["comment"], f"Faled to get parameters for group  {name}."]
            )
            return ret
        for parameter in options["parameters"].values():
            if parameter["ParameterName"] in params and params.get(
                parameter["ParameterName"]
            ) != str(parameter["ParameterValue"]):
                log.debug(
                    "Values that are being compared for %s are %s:%s.",
                    parameter["ParameterName"],
                    params.get(parameter["ParameterName"]),
                    parameter["ParameterValue"],
                )
                changed[parameter["ParameterName"]] = params.get(
                    parameter["ParameterName"]
                )
        if len(changed) > 0:
            if __opts__["test"]:
                ret["comment"] = os.linesep.join(
                    [
                        ret["comment"],
                        "Parameters {} for group {} are set to be changed.".format(
                            changed, name
                        ),
                    ]
                )
                ret["result"] = None
                return ret
            update = __salt__["boto_rds.update_parameter_group"](
                name,
                parameters=changed,
                apply_method=apply_method,
                tags=tags,
                region=region,
                key=key,
                keyid=keyid,
                profile=profile,
            )
            if "error" in update:
                ret["result"] = False
                ret["comment"] = os.linesep.join(
                    [
                        ret["comment"],
                        "Failed to change parameters {} for group {}:".format(
                            changed, name
                        ),
                        update["error"]["message"],
                    ]
                )
                return ret
            ret["changes"]["Parameters"] = changed
            ret["comment"] = os.linesep.join(
                [
                    ret["comment"],
                    f"Parameters {changed} for group {name} are changed.",
                ]
            )
        else:
            ret["comment"] = os.linesep.join(
                [
                    ret["comment"],
                    f"Parameters {params} for group {name} are present.",
                ]
            )
    return ret



"""
Verify salt-ssh stops state execution and fails with a retcode > 0
when a state rendering fails because an execution module throws an exception.
"""

import pytest

from salt.defaults.exitcodes import EX_AGGREGATE

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module", autouse=True)
def state_tree_render_module_exception(base_env_state_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_module_exception
      '127.0.0.1':
        - fail_module_exception
    """
    state_file = r"""
    This should fail being rendered:
      test.show_notification:
        - text: {{ salt["disk.usage"]("c") | yaml_dquote }}
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "fail_module_exception.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, state_tempfile:
        yield


@pytest.mark.parametrize(
    "args,retcode",
    (
        (("state.sls", "fail_module_exception"), EX_AGGREGATE),
        (("state.highstate",), EX_AGGREGATE),
        (("state.sls_id", "foo", "fail_module_exception"), EX_AGGREGATE),
        (("state.show_sls", "fail_module_exception"), EX_AGGREGATE),
        (("state.show_low_sls", "fail_module_exception"), EX_AGGREGATE),
        (("state.show_highstate",), EX_AGGREGATE),
        # state.show_lowstate exits with 0 for non-ssh as well
        (("state.show_lowstate",), 0),
        (("state.top", "top.sls"), EX_AGGREGATE),
    ),
)
def test_it(salt_ssh_cli, args, retcode):
    ret = salt_ssh_cli.run(*args)

    assert ret.returncode == retcode
    assert isinstance(ret.data, list)
    assert ret.data
    assert isinstance(ret.data[0], str)
    # While these three should usually follow each other, there
    # can be warnings in between that would break such a logic.
    assert "Rendering SLS 'base:fail_module_exception' failed:" in ret.data[0]
    assert "Problem running salt function in Jinja template:" in ret.data[0]
    assert (
        "Error running 'disk.usage': Invalid flag passed to disk.usage" in ret.data[0]
    )



"""
Simple module for creating temporary directories and files

This is a thin wrapper around Pythons tempfile module

.. versionadded:: 2015.8.0

"""

import logging
import os
import tempfile

log = logging.getLogger(__name__)


def dir(suffix="", prefix="tmp", parent=None):
    """
    Create a temporary directory

    CLI Example:

    .. code-block:: bash

        salt '*' temp.dir
        salt '*' temp.dir prefix='mytemp-' parent='/var/run/'
    """
    return tempfile.mkdtemp(suffix, prefix, parent)


def file(suffix="", prefix="tmp", parent=None):
    """
    Create a temporary file

    CLI Example:

    .. code-block:: bash

        salt '*' temp.file
        salt '*' temp.file prefix='mytemp-' parent='/var/run/'
    """
    fh_, tmp_ = tempfile.mkstemp(suffix, prefix, parent)
    os.close(fh_)
    return tmp_



import pytest

import salt.modules.freezer as freezer
from salt.exceptions import CommandExecutionError


@pytest.fixture()
def configure_loader_modules():
    return {freezer: {"__opts__": {"cachedir": ""}}}


def test_compare_no_args():
    """
    Test freezer.compare with no arguments
    """
    with pytest.raises(TypeError):
        freezer.compare()  # pylint: disable=no-value-for-parameter


def test_compare_not_enough_args():
    """
    Test freezer.compare without enough arguments
    """
    with pytest.raises(TypeError):
        freezer.compare(None)  # pylint: disable=no-value-for-parameter


def test_compare_too_many_args():
    """
    Test freezer.compare with too many arguments
    """
    with pytest.raises(TypeError):
        freezer.compare(None, None, None)  # pylint: disable=too-many-function-args


def test_compare_no_names():
    """
    Test freezer.compare with no real freeze names as arguments
    """
    with pytest.raises(CommandExecutionError):
        freezer.compare(old=None, new=None)
